/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                           BIAYSTOK 2014                               **
**                                                                        **
**                         Home Control System                            **
**                                                                        **
**                               HC_CTRL1                                 **
**                      Wersja z dnia : 2015.01.19                        **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/

/*
UART0 - RS232
UART1 - RS485
*/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include "ctrl1eeprom.h"

#define _KeyExtend_

#include "..\common\hc_keycodes.h"

#define false                          0
#define true                           1

#define LEDPort                        PORTA
#define LEDPortDirection               DDRA

#define USART0RxTxLed                  5
#define USART1RxTxLed                  6
#define KeyActiveLed                   1
#define ConnectionLed                  0
#define EEWriteBlockLed                6

#define L0LED                          0
#define L1LED                          1
#define L2LED                          2
#define L3LED                          3
#define L4LED                          4
#define L5LED                          5
#define L6LED                          6
#define L7LED                          7
#define LifeLED                        7                // PORTD.7

#define LifeLEDPort                    PORTD
#define LifeLEDDirectionPort           DDRD

#define KeybPort                       PORTC
#define KeybPortDirection              DDRC
#define InpKeybReg                     PINC

#define L0Key                          0              // PORTA . 0
#define L1Key                          1              // PORTA . 1
#define L2Key                          2              // PORTA . 2
#define L3Key                          3              // PORTA . 3
#define L4Key                          4              // PORTA . 4
#define L5Key                          5              // PORTA . 5
#define L6Key                          6              // PORTA . 6
#define L7Key                          7              // PORTA . 7


#define TimerCounterDelayLimit         3000
#define SubDevTimerCounterLimit        2048
#define QueueTimerCounterLimit         512
#define StandardLEDPulseDelay          832

#define RxTxPort                       PORTD
#define RxTxDirectionPort              DDRD
#define RxTxDirPin                     4              // PORTD.4

#define null                           0
#define cr                             0x0D
#define Space                          0x20
#define PartSeparator                  ':'
#define ParamSeparator                 '.'

#define nop() __asm__ __volatile__ ("nop")

// USART BAUD Rate Generator fosc=7.3728 MHz
#define Speed9600bps                   47
#define Speed19200bps                  23
#define Speed38400bps                  11
#define Speed115200bps                 3

#define IdleState                      0
#define BusyState                      1

#define TXD232BuffSize                 32
#define RXD232BuffSize                 16

typedef struct { volatile uint8_t  TXDCycFr ;
                 volatile uint8_t  TXDCycTo ;
                 volatile uint8_t  TXDState ;
                 volatile uint8_t  RXDCycFr ;
                 volatile uint8_t  RXDCycTo ;
                 volatile uint16_t LEDDelay ;
                 uint8_t           TXDBuffer [ TXD232BuffSize ] ;
                 uint8_t           RXDBuffer [ RXD232BuffSize ] ;
               } USART232RecT ;

#define TXD485BuffSize                 128
#define RXD485BuffSize                 16

typedef struct { volatile uint8_t  TXDCycFr ;
                 volatile uint8_t  TXDCycTo ;
                 volatile uint8_t  TXDState ;
                 volatile uint8_t  RXDCycFr ;
                 volatile uint8_t  RXDCycTo ;
                 volatile uint16_t LEDDelay ;
                 uint8_t           TXDBuffer [ TXD485BuffSize ] ;
                 uint8_t           RXDBuffer [ RXD485BuffSize ] ;
               } USART485RecT ;

#define Function0Selected              0x00
#define Function1Selected              0x01
#define Function2Selected              0x02
#define Function3Selected              0x03
#define Function4Selected              0x04
#define Function5Selected              0x05
#define Function6Selected              0x06
#define Function7Selected              0x07

#define ParamBufferSize                64
typedef uint8_t ParamBufferType [ ParamBufferSize + 1 ] ;

#define CommandLineBufferSize          126

typedef struct { uint16_t CommandLineIndex ;
                 uint8_t  CommandLine [ CommandLineBufferSize + 1 ] ;
               } CommandLineRecT ;

struct _SideServiceStruct ;

typedef struct _SideServiceStruct * PtrSideServiceType ;

typedef void ( * SideServiceOperation ) ( PtrSideServiceType /* SideService */ ) ;

typedef void ( * ServiceProcType ) ( void ) ;

typedef void ( * SendSerialProcT ) ( uint8_t /* SerialCh */ ) ;

typedef uint8_t ( * SerialDataPresentProcT ) ( void ) ;

typedef uint8_t ( * GetSerialProcT ) ( void ) ;

typedef void ( * SideOperationProcT ) ( PtrSideServiceType /* SideRecord */ ) ;

typedef struct _SideServiceStruct {
                 CommandLineRecT           CommandBuffer ;
                 SendSerialProcT           EchoSerial ;
                 SendSerialProcT           SendSerial ;
                 SerialDataPresentProcT    SerialDataPresent ;
                 GetSerialProcT            GetSerial ;
                 SideOperationProcT        SerialService ;
                 ServiceProcType           ExecuteCommand ;
                 CommandBufferType         CommandText ;
                 ParamBufferType           ParamText ;
                 DeviceIDType              DestinationDeviceName ;
                 DeviceIDType              SourceDeviceName ;
                 uint8_t                   CommandTextIndex ;
                 uint8_t                   ParamTextIndex ;
                 uint8_t                   DestDeviceNameIndex ;
                 uint8_t                   SrcDeviceNameIndex ;
               } SideServiceType ;

#define KeybDelayLimit                 2
#define KeybLongDelayLimit             25
#define ParamValueSize                 8

typedef struct {
                 uint8_t           SubDeviceLink [ ChannelsArraySize ] ;
                 uint8_t           ParamValue [ ParamValueSize + 1 ] ;
               } ParamValueDescrRecT ;

typedef enum { KeybIdle ,
               KeybWaitOnPress ,
               KeybReady ,
               KeybActive ,
               KeybWaitOnFree ,
               KeybWait
             } KeybStatusType ;

#define TimeEvent                      0x01
#define SubDeviceEvent                 0x02
#define QueueEvent                     0x04

static USART232RecT USART232 ;
static USART485RecT USART485 ;
static KeybStatusType KeybStatus ;
static uint8_t LastKeybInp ;
static uint8_t KeybCode ;
static uint16_t KeybDelayCt ;
static uint16_t KeybDelayTimer ;
static uint16_t AnyKeyDelay ;
static uint8_t LEDStatusRegister [ 2 ] ;
static uint8_t ConnectContext ;
static uint8_t CurrentTimeStatusState ;
static LEDModeSwitchT LEDModeSwitch ;
static uint16_t TimerCounter ;
volatile static uint8_t Events ;
static uint16_t SubDevTimerCounter ;
static uint8_t SubdeviceIndex ;
static uint8_t CurrentFunctionSelected ;
static SideServiceType SuperVisService ;
static SideServiceType SlaveDevService ;
static ParamValueDescrRecT ParamValueArray [ DisplParamArraySize ] ;
static uint16_t QueueTimerCounter ;


static const uint16_t CRCTable [ 16 ] = { 0x0000 ,  /* 00 */
                                          0xCC01 ,  /* 01 */
                                          0xD801 ,  /* 02 */
                                          0x1400 ,  /* 03 */
                                          0xF001 ,  /* 04 */
                                          0x3C00 ,  /* 05 */
                                          0x2800 ,  /* 06 */
                                          0xE401 ,  /* 07 */
                                          0xA001 ,  /* 08 */
                                          0x6C00 ,  /* 09 */
                                          0x7800 ,  /* 10 */
                                          0xB401 ,  /* 11 */
                                          0x5000 ,  /* 12 */
                                          0x9C01 ,  /* 13 */
                                          0x8801 ,  /* 14 */
                                          0x4400 } ;/* 15 */

static uint8_t DeviceName [ ] PROGMEM               = "HCCTRL1-control service\r" ;
static uint8_t HardwareSoftwareVersion [ ] PROGMEM  = "HVER*1_0*IX-2011.SVER*1_0*VII-2018\r" ;
static uint8_t EmptyData [ ] PROGMEM                = "<EMPTY>\r" ;
static uint8_t OK [ ] PROGMEM                       = "<OK>\r" ;
static uint8_t Fail [ ] PROGMEM                     = "<FAIL>\r" ;
static uint8_t GeneralPoolDeviceName [ ] PROGMEM    = "GENERAL" ;

static uint8_t HelloCommand [ ] PROGMEM             = "HELLO" ;
static uint8_t ConnectCommand [ ] PROGMEM           = "CONNECT" ;
static uint8_t DisConnectCommand [ ] PROGMEM        = "DISCONNECT" ;
static uint8_t SetDeviceNameCommand [ ] PROGMEM     = "SETMID" ;
static uint8_t ListSubDevCommand [ ] PROGMEM        = "LISTCH" ;
static uint8_t GetVersionCommand [ ] PROGMEM        = "GETVER" ;
static uint8_t GetIDCommand [ ] PROGMEM             = "GETID" ;
static uint8_t ReinitCommand [ ] PROGMEM            = "REINIT" ;
static uint8_t ResetEECommand [ ] PROGMEM           = "RESEE" ;
static uint8_t SaveEECommand [ ] PROGMEM            = "SAVEEE" ;
static uint8_t AddChannelCommand [ ] PROGMEM        = "ADDCH" ;
static uint8_t DelChannelCommand [ ] PROGMEM        = "DELCH" ;
static uint8_t StatusDispModeCommand [ ] PROGMEM    = "STATDM" ;
static uint8_t FunctionDispModeCommand [ ] PROGMEM  = "FUNDM" ;
static uint8_t ListKeyTableCommand [ ] PROGMEM      = "LISTKC" ;
static uint8_t AddKeyTableCommand [ ] PROGMEM       = "ADDKC" ;
static uint8_t DelKeyTableCommand [ ] PROGMEM       = "DELKC" ;
static uint8_t AddParamCommand [ ] PROGMEM          = "ADDP" ;
static uint8_t DelParamCommand [ ] PROGMEM          = "DELP" ;
static uint8_t SetStartAssocCommand [ ] PROGMEM     = "SETSA" ;
static uint8_t ResetStartAssocCommand [ ] PROGMEM   = "RESSA" ;
static uint8_t GetStartAssocCommand [ ] PROGMEM     = "GETSA" ;
static uint8_t NewPasswordCommand [ ] PROGMEM       = "NEWPW" ;
static uint8_t RunKeyCommand [ ] PROGMEM            = "RUNKC" ;
static uint8_t ParamValueCommand [ ] PROGMEM        = "RPV" ;
static uint8_t ListParamCommand [ ] PROGMEM         = "LP" ;
static uint8_t DisplayParamCommand [ ] PROGMEM      = "DP" ;


static uint8_t DisplayAsciiCommand [ ] PROGMEM      = "DA" ;
static uint8_t ClearScreenCommand [ ] PROGMEM       = "CLR" ;
static uint8_t ReplyReqCommand [ ] PROGMEM          = "REPLY" ;


typedef struct {
                 uint8_t       InpCode ;
                 uint8_t       OutCode ;
               } KeybEncoderRecT ;

#define EncodeTabeSize              36

static KeybEncoderRecT StandEncodeTabe [ EncodeTabeSize ] PROGMEM =
{
/* 00 */ { ( 1 << L0Key )                  , StandKey0Code } ,
/* 01 */ { ( 1 << L1Key )                  , StandKey1Code } ,
/* 02 */ { ( 1 << L2Key )                  , StandKey2Code } ,
/* 03 */ { ( 1 << L3Key )                  , StandKey3Code } ,
/* 04 */ { ( 1 << L4Key )                  , StandKey4Code } ,
/* 05 */ { ( 1 << L5Key )                  , StandKey5Code } ,
/* 06 */ { ( 1 << L6Key )                  , StandKey6Code } ,
/* 07 */ { ( 1 << L7Key )                  , StandKey7Code } ,
/* 08 */ { ( 1 << L0Key ) | ( 1 << L1Key ) , StandKey01Code } ,
/* 09 */ { ( 1 << L0Key ) | ( 1 << L2Key ) , StandKey02Code } ,
/* 10 */ { ( 1 << L0Key ) | ( 1 << L3Key ) , StandKey03Code } ,
/* 11 */ { ( 1 << L0Key ) | ( 1 << L4Key ) , StandKey04Code } ,
/* 12 */ { ( 1 << L0Key ) | ( 1 << L5Key ) , StandKey05Code } ,
/* 13 */ { ( 1 << L0Key ) | ( 1 << L6Key ) , StandKey06Code } ,
/* 14 */ { ( 1 << L0Key ) | ( 1 << L7Key ) , StandKey07Code } ,
/* 15 */ { ( 1 << L1Key ) | ( 1 << L2Key ) , StandKey12Code } ,
/* 16 */ { ( 1 << L1Key ) | ( 1 << L3Key ) , StandKey13Code } ,
/* 17 */ { ( 1 << L1Key ) | ( 1 << L4Key ) , StandKey14Code } ,
/* 18 */ { ( 1 << L1Key ) | ( 1 << L5Key ) , StandKey15Code } ,
/* 19 */ { ( 1 << L1Key ) | ( 1 << L6Key ) , StandKey16Code } ,
/* 20 */ { ( 1 << L1Key ) | ( 1 << L7Key ) , StandKey17Code } ,
/* 21 */ { ( 1 << L2Key ) | ( 1 << L3Key ) , StandKey23Code } ,
/* 22 */ { ( 1 << L2Key ) | ( 1 << L4Key ) , StandKey24Code } ,
/* 23 */ { ( 1 << L2Key ) | ( 1 << L5Key ) , StandKey25Code } ,
/* 24 */ { ( 1 << L2Key ) | ( 1 << L6Key ) , StandKey26Code } ,
/* 25 */ { ( 1 << L2Key ) | ( 1 << L7Key ) , StandKey27Code } ,
/* 26 */ { ( 1 << L3Key ) | ( 1 << L4Key ) , StandKey34Code } ,
/* 27 */ { ( 1 << L3Key ) | ( 1 << L5Key ) , StandKey35Code } ,
/* 28 */ { ( 1 << L3Key ) | ( 1 << L6Key ) , StandKey36Code } ,
/* 29 */ { ( 1 << L3Key ) | ( 1 << L7Key ) , StandKey37Code } ,
/* 30 */ { ( 1 << L4Key ) | ( 1 << L5Key ) , StandKey45Code } ,
/* 31 */ { ( 1 << L4Key ) | ( 1 << L6Key ) , StandKey46Code } ,
/* 32 */ { ( 1 << L4Key ) | ( 1 << L7Key ) , StandKey47Code } ,
/* 33 */ { ( 1 << L5Key ) | ( 1 << L6Key ) , StandKey56Code } ,
/* 34 */ { ( 1 << L5Key ) | ( 1 << L7Key ) , StandKey57Code } ,
/* 35 */ { ( 1 << L6Key ) | ( 1 << L7Key ) , StandKey67Code }
} ;
static KeybEncoderRecT LongEncodeTabe [ EncodeTabeSize ] PROGMEM =
{
/* 00 */ { ( 1 << L0Key )                  , LongKey0Code } ,
/* 01 */ { ( 1 << L1Key )                  , LongKey1Code } ,
/* 02 */ { ( 1 << L2Key )                  , LongKey2Code } ,
/* 03 */ { ( 1 << L3Key )                  , LongKey3Code } ,
/* 04 */ { ( 1 << L4Key )                  , LongKey4Code } ,
/* 05 */ { ( 1 << L5Key )                  , LongKey5Code } ,
/* 06 */ { ( 1 << L6Key )                  , LongKey6Code } ,
/* 07 */ { ( 1 << L7Key )                  , LongKey7Code } ,
/* 08 */ { ( 1 << L0Key ) | ( 1 << L1Key ) , LongKey01Code } ,
/* 09 */ { ( 1 << L0Key ) | ( 1 << L2Key ) , LongKey02Code } ,
/* 10 */ { ( 1 << L0Key ) | ( 1 << L3Key ) , LongKey03Code } ,
/* 11 */ { ( 1 << L0Key ) | ( 1 << L4Key ) , LongKey04Code } ,
/* 12 */ { ( 1 << L0Key ) | ( 1 << L5Key ) , LongKey05Code } ,
/* 13 */ { ( 1 << L0Key ) | ( 1 << L6Key ) , LongKey06Code } ,
/* 14 */ { ( 1 << L0Key ) | ( 1 << L7Key ) , LongKey07Code } ,
/* 15 */ { ( 1 << L1Key ) | ( 1 << L2Key ) , LongKey12Code } ,
/* 16 */ { ( 1 << L1Key ) | ( 1 << L3Key ) , LongKey13Code } ,
/* 17 */ { ( 1 << L1Key ) | ( 1 << L4Key ) , LongKey14Code } ,
/* 18 */ { ( 1 << L1Key ) | ( 1 << L5Key ) , LongKey15Code } ,
/* 19 */ { ( 1 << L1Key ) | ( 1 << L6Key ) , LongKey16Code } ,
/* 20 */ { ( 1 << L1Key ) | ( 1 << L7Key ) , LongKey17Code } ,
/* 21 */ { ( 1 << L2Key ) | ( 1 << L3Key ) , LongKey23Code } ,
/* 22 */ { ( 1 << L2Key ) | ( 1 << L4Key ) , LongKey24Code } ,
/* 23 */ { ( 1 << L2Key ) | ( 1 << L5Key ) , LongKey25Code } ,
/* 24 */ { ( 1 << L2Key ) | ( 1 << L6Key ) , LongKey26Code } ,
/* 25 */ { ( 1 << L2Key ) | ( 1 << L7Key ) , LongKey27Code } ,
/* 26 */ { ( 1 << L3Key ) | ( 1 << L4Key ) , LongKey34Code } ,
/* 27 */ { ( 1 << L3Key ) | ( 1 << L5Key ) , LongKey35Code } ,
/* 28 */ { ( 1 << L3Key ) | ( 1 << L6Key ) , LongKey36Code } ,
/* 29 */ { ( 1 << L3Key ) | ( 1 << L7Key ) , LongKey37Code } ,
/* 30 */ { ( 1 << L4Key ) | ( 1 << L5Key ) , LongKey45Code } ,
/* 31 */ { ( 1 << L4Key ) | ( 1 << L6Key ) , LongKey46Code } ,
/* 32 */ { ( 1 << L4Key ) | ( 1 << L7Key ) , LongKey47Code } ,
/* 33 */ { ( 1 << L5Key ) | ( 1 << L6Key ) , LongKey56Code } ,
/* 34 */ { ( 1 << L5Key ) | ( 1 << L7Key ) , LongKey57Code } ,
/* 35 */ { ( 1 << L6Key ) | ( 1 << L7Key ) , LongKey67Code }
} ;


typedef struct {
                 uint8_t             CodeElement ;
                 uint8_t *           NameElement ;
               } CodeAndNameAssociationRecT ;

#define KeyCodeAndNameAssociationTabeSize            73

static CodeAndNameAssociationRecT KeyCodeNameAssociationTabe [ KeyCodeAndNameAssociationTabeSize ] PROGMEM =
{
/* 00 */ { NoneKeyCode      , NoneKeyCodeName    } ,
/* 01 */ { StandKey0Code    , StandKey0CodeName  } ,
/* 02 */ { StandKey1Code    , StandKey1CodeName  } ,
/* 03 */ { StandKey2Code    , StandKey2CodeName  } ,
/* 04 */ { StandKey3Code    , StandKey3CodeName  } ,
/* 05 */ { StandKey4Code    , StandKey4CodeName  } ,
/* 06 */ { StandKey5Code    , StandKey5CodeName  } ,
/* 07 */ { StandKey6Code    , StandKey6CodeName  } ,
/* 08 */ { StandKey7Code    , StandKey7CodeName  } ,
/* 09 */ { StandKey01Code   , StandKey01CodeName } ,
/* 10 */ { StandKey02Code   , StandKey02CodeName } ,
/* 11 */ { StandKey03Code   , StandKey03CodeName } ,
/* 12 */ { StandKey04Code   , StandKey04CodeName } ,
/* 13 */ { StandKey05Code   , StandKey05CodeName } ,
/* 14 */ { StandKey06Code   , StandKey06CodeName } ,
/* 15 */ { StandKey07Code   , StandKey07CodeName } ,
/* 16 */ { StandKey12Code   , StandKey12CodeName } ,
/* 17 */ { StandKey13Code   , StandKey13CodeName } ,
/* 18 */ { StandKey14Code   , StandKey14CodeName } ,
/* 19 */ { StandKey15Code   , StandKey15CodeName } ,
/* 10 */ { StandKey16Code   , StandKey16CodeName } ,
/* 21 */ { StandKey17Code   , StandKey17CodeName } ,
/* 22 */ { StandKey23Code   , StandKey23CodeName } ,
/* 23 */ { StandKey24Code   , StandKey24CodeName } ,
/* 24 */ { StandKey25Code   , StandKey25CodeName } ,
/* 25 */ { StandKey26Code   , StandKey26CodeName } ,
/* 26 */ { StandKey27Code   , StandKey27CodeName } ,
/* 27 */ { StandKey34Code   , StandKey34CodeName } ,
/* 28 */ { StandKey35Code   , StandKey35CodeName } ,
/* 29 */ { StandKey36Code   , StandKey36CodeName } ,
/* 30 */ { StandKey37Code   , StandKey37CodeName } ,
/* 31 */ { StandKey45Code   , StandKey45CodeName } ,
/* 32 */ { StandKey46Code   , StandKey46CodeName } ,
/* 33 */ { StandKey47Code   , StandKey47CodeName } ,
/* 34 */ { StandKey56Code   , StandKey56CodeName } ,
/* 35 */ { StandKey57Code   , StandKey57CodeName } ,
/* 36 */ { StandKey67Code   , StandKey67CodeName } ,
/* 37 */ { LongKey0Code     , LongKey0CodeName   } ,
/* 38 */ { LongKey1Code     , LongKey1CodeName   } ,
/* 39 */ { LongKey2Code     , LongKey2CodeName   } ,
/* 40 */ { LongKey3Code     , LongKey3CodeName   } ,
/* 41 */ { LongKey4Code     , LongKey4CodeName   } ,
/* 42 */ { LongKey5Code     , LongKey5CodeName   } ,
/* 43 */ { LongKey6Code     , LongKey6CodeName   } ,
/* 44 */ { LongKey7Code     , LongKey7CodeName   } ,
/* 45 */ { LongKey01Code    , LongKey01CodeName  } ,
/* 46 */ { LongKey02Code    , LongKey02CodeName  } ,
/* 47 */ { LongKey03Code    , LongKey03CodeName  } ,
/* 48 */ { LongKey04Code    , LongKey04CodeName  } ,
/* 49 */ { LongKey05Code    , LongKey05CodeName  } ,
/* 50 */ { LongKey06Code    , LongKey06CodeName  } ,
/* 51 */ { LongKey07Code    , LongKey07CodeName  } ,
/* 52 */ { LongKey12Code    , LongKey12CodeName  } ,
/* 53 */ { LongKey13Code    , LongKey13CodeName  } ,
/* 54 */ { LongKey14Code    , LongKey14CodeName  } ,
/* 55 */ { LongKey15Code    , LongKey15CodeName  } ,
/* 56 */ { LongKey16Code    , LongKey16CodeName  } ,
/* 57 */ { LongKey17Code    , LongKey17CodeName  } ,
/* 58 */ { LongKey23Code    , LongKey23CodeName  } ,
/* 59 */ { LongKey24Code    , LongKey24CodeName  } ,
/* 60 */ { LongKey25Code    , LongKey25CodeName  } ,
/* 61 */ { LongKey26Code    , LongKey26CodeName  } ,
/* 62 */ { LongKey27Code    , LongKey27CodeName  } ,
/* 63 */ { LongKey34Code    , LongKey34CodeName  } ,
/* 64 */ { LongKey35Code    , LongKey35CodeName  } ,
/* 65 */ { LongKey36Code    , LongKey36CodeName  } ,
/* 66 */ { LongKey37Code    , LongKey37CodeName  } ,
/* 67 */ { LongKey45Code    , LongKey45CodeName  } ,
/* 68 */ { LongKey46Code    , LongKey46CodeName  } ,
/* 69 */ { LongKey47Code    , LongKey47CodeName  } ,
/* 70 */ { LongKey56Code    , LongKey56CodeName  } ,
/* 71 */ { LongKey57Code    , LongKey57CodeName  } ,
/* 72 */ { LongKey67Code    , LongKey67CodeName  }
} ;


typedef void ( * CommandExecuteServiceProcT ) ( PtrSideServiceType SideData ) ;
typedef struct {
                 CommandExecuteServiceProcT  CommandService ;
                 uint8_t *                   CommandName ;
               } CommandNameAssociationRecT ;

static void ConnectCommandExecute ( PtrSideServiceType SideData ) ;
static void DisConnectCommandExecute ( PtrSideServiceType SideData ) ;
static void HelloCommandExecute ( PtrSideServiceType SideData ) ;
static void ListSubDevCommandExecute ( PtrSideServiceType SideData ) ;
static void GetVersionCommandExecute ( PtrSideServiceType SideData ) ;
static void GetIDCommandExecute ( PtrSideServiceType SideData ) ;
static void ReinitCommandExecute ( PtrSideServiceType SideData ) ;
static void ResetEECommandExecute ( PtrSideServiceType SideData ) ;
static void SaveEECommandExecute ( PtrSideServiceType SideData ) ;
static void ListParamCommandExecute ( PtrSideServiceType SideData ) ;
static void DisplayParamCommandExecute ( PtrSideServiceType SideData ) ;
static void AddChannelCommandExecute ( PtrSideServiceType SideData ) ;
static void DelChannelCommandExecute ( PtrSideServiceType SideData ) ;
static void SetDeviceNameCommandExecute ( PtrSideServiceType SideData ) ;
static void StatusDispModeCommandExecute ( PtrSideServiceType SideData ) ;
static void FunctionDispModeCommandExecute ( PtrSideServiceType SideData ) ;
static void AddParamCommandExecute ( PtrSideServiceType SideData ) ;
static void DelParamCommandExecute ( PtrSideServiceType SideData ) ;
static void SetStartAssocCommandExecute ( PtrSideServiceType SideData ) ;
static void ResetStartAssocCommandExecute ( PtrSideServiceType SideData ) ;
static void GetStartAssocCommandExecute ( PtrSideServiceType SideData ) ;
static void ListKeyTableCommandExecute ( PtrSideServiceType SideData ) ;
static void AddKeyTableCommandExecute ( PtrSideServiceType SideData ) ;
static void DelKeyTableCommandExecute ( PtrSideServiceType SideData ) ;
static void NewPasswordCommandExecute ( PtrSideServiceType SideData ) ;
static void RunKeyCommandExecute ( PtrSideServiceType SideData ) ;
static void ParamValueCommandExecute ( PtrSideServiceType SideData ) ;


#define SvCommandNameAssociationTabeSize             27

static CommandNameAssociationRecT SvCommandNameAssociationTabe [ SvCommandNameAssociationTabeSize ] PROGMEM =
{
/* 00 */ { ConnectCommandExecute               , ConnectCommand } ,
/* 01 */ { DisConnectCommandExecute            , DisConnectCommand } ,
/* 02 */ { SetDeviceNameCommandExecute         , SetDeviceNameCommand } ,
/* 03 */ { HelloCommandExecute                 , HelloCommand } ,
/* 04 */ { ListSubDevCommandExecute            , ListSubDevCommand } ,
/* 05 */ { GetVersionCommandExecute            , GetVersionCommand } ,
/* 06 */ { GetIDCommandExecute                 , GetIDCommand } ,
/* 07 */ { ReinitCommandExecute                , ReinitCommand } ,
/* 08 */ { ResetEECommandExecute               , ResetEECommand } ,
/* 09 */ { SaveEECommandExecute                , SaveEECommand } ,
/* 10 */ { ListParamCommandExecute             , ListParamCommand } ,
/* 11 */ { DisplayParamCommandExecute          , DisplayParamCommand } ,
/* 12 */ { AddChannelCommandExecute            , AddChannelCommand } ,
/* 13 */ { DelChannelCommandExecute            , DelChannelCommand } ,
/* 14 */ { StatusDispModeCommandExecute        , StatusDispModeCommand } ,
/* 15 */ { FunctionDispModeCommandExecute      , FunctionDispModeCommand } ,
/* 16 */ { AddParamCommandExecute              , AddParamCommand } ,
/* 17 */ { DelParamCommandExecute              , DelParamCommand } ,
/* 18 */ { SetStartAssocCommandExecute         , SetStartAssocCommand } ,
/* 19 */ { ResetStartAssocCommandExecute       , ResetStartAssocCommand } ,
/* 20 */ { GetStartAssocCommandExecute         , GetStartAssocCommand } ,
/* 21 */ { ListKeyTableCommandExecute          , ListKeyTableCommand } ,
/* 22 */ { AddKeyTableCommandExecute           , AddKeyTableCommand } ,
/* 23 */ { DelKeyTableCommandExecute           , DelKeyTableCommand } ,
/* 24 */ { NewPasswordCommandExecute           , NewPasswordCommand } ,
/* 25 */ { RunKeyCommandExecute                , RunKeyCommand } ,
/* 26 */ { ParamValueCommandExecute            , ParamValueCommand } ,
} ;



static uint8_t HexalTable [ ] PROGMEM                = "0123456789ABCDEF" ;

typedef enum { SuperVRxTxLED ,
               SubDevRxTxLED ,
               AnyKeyLED ,
               ConnectLED ,
               EEWriteLED ,
               Key0LED ,
               Key1LED ,
               Key2LED ,
               Key3LED ,
               Key4LED ,
               Key5LED ,
               Key6LED ,
               Key7LED ,
             } LEDSinalT ;


static void LEDOn ( LEDSinalT LEDSignal )
{
  /*----------------------------------------------------------------*/
  switch ( LEDSignal )
  {
    case SuperVRxTxLED :
      LEDStatusRegister [ 0 ] &= ~ ( 1 << USART0RxTxLed ) ;
      break ;
    case SubDevRxTxLED :
      LEDStatusRegister [ 0 ] &= ~ ( 1 << USART1RxTxLed ) ;
      break ;
    case AnyKeyLED :
      LEDStatusRegister [ 0 ] &= ~ ( 1 << KeyActiveLed ) ;
      break ;
    case ConnectLED :
      LEDStatusRegister [ 0 ] &= ~ ( 1 << ConnectionLed ) ;
      break ;
    case EEWriteLED :
      LEDStatusRegister [ 0 ] &= ~ ( 1 << EEWriteBlockLed ) ;
      break ;
    case Key0LED :
      LEDStatusRegister [ 1 ] &= ~ ( 1 << L0LED ) ;
      break ;
    case Key1LED :
      LEDStatusRegister [ 1 ] &= ~ ( 1 << L1LED ) ;
      break ;
    case Key2LED :
      LEDStatusRegister [ 1 ] &= ~ ( 1 << L2LED ) ;
      break ;
    case Key3LED :
      LEDStatusRegister [ 1 ] &= ~ ( 1 << L3LED ) ;
      break ;
    case Key4LED :
      LEDStatusRegister [ 1 ] &= ~ ( 1 << L4LED ) ;
      break ;
    case Key5LED :
      LEDStatusRegister [ 1 ] &= ~ ( 1 << L5LED ) ;
      break ;
    case Key6LED :
      LEDStatusRegister [ 1 ] &= ~ ( 1 << L6LED ) ;
      break ;
    case Key7LED :
      LEDStatusRegister [ 1 ] &= ~ ( 1 << L7LED ) ;
      break ;
  } /* switch */ ;
  switch ( LEDModeSwitch )
  {
    case FunctionDispMode :
      LEDPort = LEDStatusRegister [ 1 ] ;
      break ;
    case StatusDispMode :
      LEDPort = LEDStatusRegister [ 0 ] ;
      break ;
  } /* switch */ ;
} /* LEDOn */


static void LEDOff ( LEDSinalT LEDSignal )
{
  /*----------------------------------------------------------------*/
  switch ( LEDSignal )
  {
    case SuperVRxTxLED :
      LEDStatusRegister [ 0 ] |= ( 1 << USART0RxTxLed ) ;
      break ;
    case SubDevRxTxLED :
      LEDStatusRegister [ 0 ] |= ( 1 << USART1RxTxLed ) ;
      break ;
    case AnyKeyLED :
      LEDStatusRegister [ 0 ] |= ( 1 << KeyActiveLed ) ;
      break ;
    case ConnectLED :
      LEDStatusRegister [ 0 ] |= ( 1 << ConnectionLed ) ;
      break ;
    case EEWriteLED :
      LEDStatusRegister [ 0 ] |= ( 1 << EEWriteBlockLed ) ;
      break ;
    case Key0LED :
      LEDStatusRegister [ 1 ] |= ( 1 << L0LED ) ;
      break ;
    case Key1LED :
      LEDStatusRegister [ 1 ] |= ( 1 << L1LED ) ;
      break ;
    case Key2LED :
      LEDStatusRegister [ 1 ] |= ( 1 << L2LED ) ;
      break ;
    case Key3LED :
      LEDStatusRegister [ 1 ] |= ( 1 << L3LED ) ;
      break ;
    case Key4LED :
      LEDStatusRegister [ 1 ] |= ( 1 << L4LED ) ;
      break ;
    case Key5LED :
      LEDStatusRegister [ 1 ] |= ( 1 << L5LED ) ;
      break ;
    case Key6LED :
      LEDStatusRegister [ 1 ] |= ( 1 << L6LED ) ;
      break ;
    case Key7LED :
      LEDStatusRegister [ 1 ] |= ( 1 << L7LED ) ;
      break ;
  } /* switch */ ;
  switch ( LEDModeSwitch )
  {
    case FunctionDispMode :
      LEDPort = LEDStatusRegister [ 1 ] ;
      break ;
    case StatusDispMode :
      LEDPort = LEDStatusRegister [ 0 ] ;
      break ;
  } /* switch */ ;
} /* LEDOff */


static void StartUSART232RxTxLED ( void )
{
  /*----------------------------------------------------------------*/
  LEDOn ( SuperVRxTxLED ) ;
  USART232 . LEDDelay = StandardLEDPulseDelay ;
} /* StartUSART232RxTxLED */


static void StopUSART232RxTxLED ( void )
{
  /*----------------------------------------------------------------*/
  LEDOff ( SuperVRxTxLED ) ;
} /* StopUSART232RxTxLED */


static void StartUSART485RxTxLED ( void )
{
  /*----------------------------------------------------------------*/
  LEDOn ( SubDevRxTxLED ) ;
  USART485 . LEDDelay = StandardLEDPulseDelay ;
} /* StartUSART485RxTxLED */


static void StopUSART485RxTxLED ( void )
{
  /*----------------------------------------------------------------*/
  LEDOff ( SubDevRxTxLED ) ;
} /* StopUSART485RxTxLED */


static void SetFunctionIndicator ( uint8_t FunctionSelected )
{
  /*----------------------------------------------------------------*/
  switch ( FunctionSelected )
  {
    case Function0Selected :
      LEDOn ( Key0LED ) ;
      break ;
    case Function1Selected :
      LEDOn ( Key1LED ) ;
      break ;
    case Function2Selected :
      LEDOn ( Key2LED ) ;
      break ;
    case Function3Selected :
      LEDOn ( Key3LED ) ;
      break ;
    case Function4Selected :
      LEDOn ( Key4LED ) ;
      break ;
    case Function5Selected :
      LEDOn ( Key5LED ) ;
      break ;
    case Function6Selected :
      LEDOn ( Key6LED ) ;
      break ;
    case Function7Selected :
      LEDOn ( Key7LED ) ;
      break ;
  } /* switch */ ;
} /* SetFunctionIndicator */


static void ClearFunctionIndicator ( uint8_t FunctionSelected )
{
  /*----------------------------------------------------------------*/
  switch ( FunctionSelected )
  {
    case Function0Selected :
      LEDOff ( Key0LED ) ;
      break ;
    case Function1Selected :
      LEDOff ( Key1LED ) ;
      break ;
    case Function2Selected :
      LEDOff ( Key2LED ) ;
      break ;
    case Function3Selected :
      LEDOff ( Key3LED ) ;
      break ;
    case Function4Selected :
      LEDOff ( Key4LED ) ;
      break ;
    case Function5Selected :
      LEDOff ( Key5LED ) ;
      break ;
    case Function6Selected :
      LEDOff ( Key6LED ) ;
      break ;
    case Function7Selected :
      LEDOff ( Key7LED ) ;
      break ;
  } /* switch */ ;
} /* ClearFunctionIndicator */


static uint8_t KeyPressed ( void )
{
  /*----------------------------------------------------------------*/
  return ( KeybCode != 0 ) ;
} /* KeyPressed */


static uint8_t ReadKey ( void )
{
  uint8_t TempCh ;
  /*----------------------------------------------------------------*/
  TempCh = KeybCode ;
  KeybCode = 0 ;
  return ( TempCh ) ;
} /* ReadKey */


static void StoreKey ( uint8_t NewCh )
{
  /*----------------------------------------------------------------*/
  AnyKeyDelay = StandardLEDPulseDelay ;
  LEDOn ( AnyKeyLED ) ;
  KeybCode = NewCh ;
} /* StoreKey */


static uint8_t HardwareKeybRead ( void )
{
  uint8_t KeybData ;
  uint8_t Mask ;
  /*----------------------------------------------------------------*/
  KeybData = InpKeybReg ;
  Mask = ( 1 << L0Key ) | ( 1 << L1Key ) |
         ( 1 << L2Key ) | ( 1 << L3Key ) |
         ( 1 << L4Key ) | ( 1 << L5Key ) |
         ( 1 << L6Key ) | ( 1 << L7Key ) ;
  return ( KeybData & Mask ) ;
} /* HardwareKeybRead */


static uint8_t AnyKeyActive ( void )
{
  uint8_t KeybData ;
  uint8_t Mask ;
  /*----------------------------------------------------------------*/
  Mask = ( 1 << L0Key ) | ( 1 << L1Key ) |
         ( 1 << L2Key ) | ( 1 << L3Key ) |
         ( 1 << L4Key ) | ( 1 << L5Key ) |
         ( 1 << L6Key ) | ( 1 << L7Key ) ;
  KeybData = HardwareKeybRead ( ) ;
  if ( KeybData == Mask )
    return ( false ) ;
  else
    return ( true ) ;
} /* AnyKeyActive */


static void DecodeKey ( uint8_t Kbd ,
                        uint8_t LongDelayMode )
{
  uint8_t Loop ;
  uint8_t InpC ;
  uint8_t OutC ;
  uint16_t MemAddress ;
  /*----------------------------------------------------------------*/
  if ( LongDelayMode )
    MemAddress = ( uint16_t ) LongEncodeTabe ;
  else
    MemAddress = ( uint16_t ) StandEncodeTabe ;
  for ( Loop = 0 ; Loop < EncodeTabeSize ; Loop ++ )
  {
    InpC = pgm_read_byte_near ( MemAddress ++ ) ;
    OutC = pgm_read_byte_near ( MemAddress ++ ) ;
    if ( Kbd == InpC )
    {
      StoreKey ( OutC ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
} /* DecodeKey */


static void KeybService ( void )
{
  uint8_t Mask ;
  /*----------------------------------------------------------------*/
  switch ( KeybStatus )
  {
    case KeybIdle        :
      if ( AnyKeyActive ( ) )
      {
        KeybStatus = KeybWaitOnPress ;
        KeybDelayCt = KeybDelayLimit ;
      } /* if */ ;
      break ;
    case KeybWaitOnPress :
      KeybDelayCt -- ;
      if ( ! KeybDelayCt )
      {
        if ( AnyKeyActive ( ) )
          KeybStatus = KeybReady ;
        else
          KeybStatus = KeybIdle ;
      } /* if */ ;
      break ;
    case KeybReady       :
      Mask = ( 1 << L0Key ) | ( 1 << L1Key ) |
             ( 1 << L2Key ) | ( 1 << L3Key ) |
             ( 1 << L4Key ) | ( 1 << L5Key ) |
             ( 1 << L6Key ) | ( 1 << L7Key ) ;
      LastKeybInp = ( HardwareKeybRead ( ) ^ 0xFF ) & Mask ;
      KeybDelayCt = KeybLongDelayLimit ;
      KeybStatus = KeybActive ;
      AnyKeyDelay = StandardLEDPulseDelay ;
      break ;
    case KeybActive      :
      KeybDelayCt -- ;
      if ( ! KeybDelayCt )
      {
        DecodeKey ( LastKeybInp , true ) ;
        KeybStatus = KeybWaitOnFree ;
        KeybDelayCt = KeybDelayLimit ;
      } /* if ... */
      else
      {
        if ( ! AnyKeyActive ( ) )
        {
          DecodeKey ( LastKeybInp , false ) ;
          KeybStatus = KeybWaitOnFree ;
          KeybDelayCt = KeybDelayLimit ;
        } /* if */ ;
      } /* if ... else */ ;
      break ;
    case KeybWaitOnFree  :
      KeybDelayCt -- ;
      if ( ! KeybDelayCt )
      {
        KeybStatus = KeybWait ;
      } /* if */ ;
      break ;
    case KeybWait        :
      if ( ! AnyKeyActive ( ) )
      {
        KeybStatus = KeybIdle ;
      } /* if */ ;
      break ;
  } /* switch */ ;
} /* KeybService */


ISR ( TIMER0_OVF_vect )
{
  /*----------------------------------------------------------------*/
  TimerCounter ++ ;
  if ( TimerCounter > TimerCounterDelayLimit )
  {
    TimerCounter = 0 ;
    Events |= TimeEvent ;
  } /* if */ ;
  if ( USART232 . LEDDelay )
  {
    USART232 . LEDDelay -- ;
    if ( ! USART232 . LEDDelay )
    {
      StopUSART232RxTxLED ( ) ;
    } /* if */ ;
  } /* if */ ;
  if ( USART485 . LEDDelay )
  {
    USART485 . LEDDelay -- ;
    if ( ! USART485 . LEDDelay )
    {
      StopUSART485RxTxLED ( ) ;
    } /* if */ ;
  } /* if */ ;
  SubDevTimerCounter ++ ;
  if ( SubDevTimerCounter > SubDevTimerCounterLimit )
  {
    SubDevTimerCounter = 0 ;
    Events |= SubDeviceEvent ;
  } /* if */ ;
  QueueTimerCounter ++ ;
  if ( QueueTimerCounter > QueueTimerCounterLimit )
  {
    QueueTimerCounter = 0 ;
    Events |= QueueEvent ;
  } /* if */ ;
  if ( AnyKeyDelay )
  {
    AnyKeyDelay -- ;
    if ( ! AnyKeyDelay )
    {
      LEDOff ( AnyKeyLED ) ;
    } /* if */ ;
  } /* if */ ;
  KeybDelayTimer ++ ;
  if ( KeybDelayTimer > 100 )
  {
    KeybDelayTimer = 0 ;
    KeybService ( ) ;
  } /* if */ ;
} /* TIMER0_OVF_vect */


static void USART485ReceiveMode ( void )
{
  /*----------------------------------------------------------------*/
  RxTxPort &= ~ ( 1 << RxTxDirPin ) ;
} /* USART485ReceiveMode */


static void USART485TransmitMode ( void )
{
  /*----------------------------------------------------------------*/
  RxTxPort |= ( 1 << RxTxDirPin ) ;
} /* USART485TransmitMode */


ISR ( USART0_RX_vect )
{
  uint8_t RecCh ;
  /*----------------------------------------------------------------*/
  RecCh = UDR0 ;
  USART232 . RXDBuffer [ USART232 . RXDCycTo ] = RecCh ;
  USART232 . RXDCycTo ++ ;
  if ( USART232 . RXDCycTo >= RXD232BuffSize )
    USART232 . RXDCycTo = 0 ;
  StartUSART232RxTxLED ( ) ;
} /* USART0_RX_vect */


ISR ( USART0_TX_vect )
{
  /*----------------------------------------------------------------*/
  if ( USART232 . TXDCycFr == USART232 . TXDCycTo )
  {
    USART232 . TXDState = IdleState ;
  } /* if ... */
  else
  {
    UDR0 = USART232 . TXDBuffer [ USART232 . TXDCycFr ] ;
    USART232 . TXDCycFr ++ ;
    if ( USART232 . TXDCycFr >= TXD232BuffSize )
      USART232 . TXDCycFr = 0 ;
    StartUSART232RxTxLED ( ) ;
  } /* if ... else */ ;
} /* USART0_TX_vect */


ISR ( USART1_RX_vect )
{
  uint8_t RecCh ;
  /*----------------------------------------------------------------*/
  RecCh = UDR1 ;
  USART485 . RXDBuffer [ USART485 . RXDCycTo ] = RecCh ;
  USART485 . RXDCycTo ++ ;
  if ( USART485 . RXDCycTo >= RXD485BuffSize )
    USART485 . RXDCycTo = 0 ;
  StartUSART485RxTxLED ( ) ;
} /* USART1_RX_vect */


ISR ( USART1_TX_vect )
{
  /*----------------------------------------------------------------*/
  if ( USART485 . TXDCycFr == USART485 . TXDCycTo )
  {
    USART485 . TXDState = IdleState ;
    USART485ReceiveMode ( ) ;
  } /* if ... */
  else
  {
    UDR1 = USART485 . TXDBuffer [ USART485 . TXDCycFr ] ;
    USART485 . TXDCycFr ++ ;
    if ( USART485 . TXDCycFr >= TXD485BuffSize )
      USART485 . TXDCycFr = 0 ;
    StartUSART485RxTxLED ( ) ;
  } /* if ... else */ ;
} /* USART1_TX_vect */


static void USART232SendSerial ( uint8_t SerialCh )
{
  uint8_t NextInx ;
  /*----------------------------------------------------------------*/
  cli ( ) ;
  StartUSART232RxTxLED ( ) ;
  if ( USART232 . TXDState == IdleState )
  {
    USART232 . TXDState = BusyState ;
    UDR0 = SerialCh ;
    sei ( ) ;
  } /* if ... */
  else
  {
    for ( ; ; )
    {
      NextInx = USART232 . TXDCycTo + 1 ;
      if ( NextInx >= TXD232BuffSize )
        NextInx = 0 ;
      if ( NextInx != USART232 . TXDCycFr )
        break ;
      sei ( ) ;
      nop ( ) ;
      nop ( ) ;
      nop ( ) ;
      wdt_reset ( ) ;
      cli ( ) ;
    } /* for */ ;
    USART232 . TXDBuffer [ USART232 . TXDCycTo ] = SerialCh ;
    USART232 . TXDCycTo ++ ;
    if ( USART232 . TXDCycTo >= TXD232BuffSize )
      USART232 . TXDCycTo = 0 ;
    sei ( ) ;
  } /* if ... else */ ;
} /* USART232SendSerial */


static uint8_t USART232DataPresent ( void )
{
  /*----------------------------------------------------------------*/
  if ( USART232 . RXDCycFr == USART232 . RXDCycTo )
    return ( false ) ;
  else
    return ( true ) ;
} /* USART232DataPresent */


static uint8_t USART232GetSerial ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = USART232 . RXDBuffer [ USART232 . RXDCycFr ] ;
  USART232 . RXDCycFr ++ ;
  if ( USART232 . RXDCycFr >= RXD232BuffSize )
    USART232 . RXDCycFr = 0 ;
  return ( Ch ) ;
} /* USART232GetSerial */


static void USART485SendSerial ( uint8_t SerialCh )
{
  uint8_t NextInx ;
  /*----------------------------------------------------------------*/
  cli ( ) ;
  StartUSART485RxTxLED ( ) ;
  if ( USART485 . TXDState == IdleState )
  {
    USART485TransmitMode ( ) ;
    USART485 . TXDState = BusyState ;
    UDR1 = SerialCh ;
    sei ( ) ;
  } /* if ... */
  else
  {
    for ( ; ; )
    {
      NextInx = USART485 . TXDCycTo + 1 ;
      if ( NextInx >= TXD485BuffSize )
        NextInx = 0 ;
      if ( NextInx != USART485 . TXDCycFr )
        break ;
      sei ( ) ;
      nop ( ) ;
      nop ( ) ;
      nop ( ) ;
      wdt_reset ( ) ;
      cli ( ) ;
    } /* for */ ;
    USART485 . TXDBuffer [ USART485 . TXDCycTo ] = SerialCh ;
    USART485 . TXDCycTo ++ ;
    if ( USART485 . TXDCycTo >= TXD485BuffSize )
      USART485 . TXDCycTo = 0 ;
    sei ( ) ;
  } /* if ... else */ ;
} /* USART485SendSerial */


static uint8_t USART485DataPresent ( void )
{
  /*----------------------------------------------------------------*/
  if ( USART485 . RXDCycFr == USART485 . RXDCycTo )
    return ( false ) ;
  else
    return ( true ) ;
} /* USART485DataPresent */


static uint8_t USART485GetSerial ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = USART485 . RXDBuffer [ USART485 . RXDCycFr ] ;
  USART485 . RXDCycFr ++ ;
  if ( USART485 . RXDCycFr >= RXD485BuffSize )
    USART485 . RXDCycFr = 0 ;
  return ( Ch ) ;
} /* USART485DataPresent */


static void USART0HardwareInit ( uint16_t SvSerialSpeed )
{
  /*----------------------------------------------------------------*/
  UBRR0H = ( uint8_t ) ( SvSerialSpeed >> 8 ) ;
  UBRR0L = ( uint8_t ) ( SvSerialSpeed & 0xFF ) ;
  UCSR0A = 0 ;
  UCSR0B = ( 1 << RXCIE0 ) | ( 1 << TXCIE0 ) | ( 1 << RXEN0 ) | ( 1 << TXEN0 ) ;
  UCSR0C = ( 1 << UCSZ01 ) | ( 1 << UCSZ00 ) ;
} /* USART0HardwareInit */


static void USART1HardwareInit ( uint16_t PWCSerialSpeed )
{
  /*----------------------------------------------------------------*/
  UBRR1H = ( uint8_t ) ( PWCSerialSpeed >> 8 ) ;
  UBRR1L = ( uint8_t ) ( PWCSerialSpeed & 0xFF ) ;
  UCSR1A = 0 ;
  UCSR1B = ( 1 << RXCIE1 ) | ( 1 << TXCIE1 ) | ( 1 << RXEN1 ) | ( 1 << TXEN1 ) ;
  UCSR1C = ( 1 << UCSZ11 ) | ( 1 << UCSZ10 ) ;
} /* USART1HardwareInit */


static void USART232Reset ( void )
{
  /*----------------------------------------------------------------*/
  USART232 . RXDCycFr = 0 ;
  USART232 . RXDCycTo = 0 ;
  USART232 . TXDCycFr = 0 ;
  USART232 . TXDCycTo = 0 ;
  USART232 . TXDState = IdleState ;
  USART232 . LEDDelay = 0 ;
} /* USART232Reset */


static void USART485Reset ( void )
{
  /*----------------------------------------------------------------*/
  USART485 . RXDCycFr = 0 ;
  USART485 . RXDCycTo = 0 ;
  USART485 . TXDCycFr = 0 ;
  USART485 . TXDCycTo = 0 ;
  USART485 . TXDState = IdleState ;
  USART485 . LEDDelay = 0 ;
} /* USART485Reset */


static void HardwareInit ( uint16_t PWCSerialSpeed ,
                           uint16_t SvSerialSpeed )
{
  /*----------------------------------------------------------------*/
  TCCR0B = ( 1 << CS01 ) ;                                 // CS02=0 CS01=1 CS00=0 : F clk div 8
  TIMSK0 = ( 1 << TOIE0 ) ;                                // TOIE0: Timer/Counter0 Overflow Interrupt Enable
  USART0HardwareInit ( SvSerialSpeed ) ;
  USART1HardwareInit ( PWCSerialSpeed ) ;
} /* HardwareInit */


static void EnvirInit ( void )
{
  uint8_t Bits ;
  /*----------------------------------------------------------------*/
  Bits = 0xFF ;
  LEDPortDirection = Bits ;
  LEDPort = Bits ;
  LEDStatusRegister [ 0 ] = Bits ;
  LEDStatusRegister [ 1 ] = Bits ;
  Bits = 0xFF ;
  Bits ^= ( 1 << L0Key ) | ( 1 << L1Key ) |
          ( 1 << L2Key ) | ( 1 << L3Key ) |
          ( 1 << L4Key ) | ( 1 << L5Key ) |
          ( 1 << L6Key ) | ( 1 << L7Key ) ;
  KeybPortDirection = Bits ;
  KeybPort = 0 ;
  Bits = ( 1 << RxTxDirPin ) ;
  RxTxDirectionPort |= Bits ;
  USART485ReceiveMode ( ) ;
  LifeLEDDirectionPort |= ( 1 ) << LifeLED ;
  LifeLEDPort |= ( 1 ) << LifeLED ;
} /* EnvirInit */


static void Send ( PtrSideServiceType SideData ,
                   uint8_t            Ch )
{
  /*----------------------------------------------------------------*/
  if ( SideData -> CommandBuffer . CommandLineIndex < CommandLineBufferSize )
  {
    SideData -> CommandBuffer . CommandLine [ SideData -> CommandBuffer . CommandLineIndex ++ ] = Ch ;
    SideData -> CommandBuffer . CommandLine [ SideData -> CommandBuffer . CommandLineIndex ] = 0 ;
  } /* if */ ;
} /* Send */


static void SendTxt ( PtrSideServiceType SideData ,
                      uint8_t *          String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = * String ++ ;
    if ( Data )
      Send ( SideData , Data ) ;
    else
      break ;
  } /* for */ ;
} /* SendTxt */


static void SendFlashTxt ( PtrSideServiceType SideData ,
                           uint16_t           String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = pgm_read_byte_near ( String ) ;
    if ( Data )
      Send ( SideData , Data ) ;
    else
      break ;
    String ++ ;
  } /* for */ ;
} /* SendFlashTxt */


static void FlushData ( PtrSideServiceType SideData )
{
  uint8_t Ch ;
  uint16_t Loop ;
  /*----------------------------------------------------------------*/
  if ( SideData -> CommandBuffer . CommandLineIndex )
  {
    for ( Loop = 0 ; Loop < SideData -> CommandBuffer . CommandLineIndex ; Loop ++ )
    {
      Ch = SideData -> CommandBuffer . CommandLine [ Loop ] ;
      SideData -> SendSerial ( Ch ) ;
    } /* for */ ;
  } /* if */ ;
} /* FlushData */


static void SuperVFlushData ( void )
{
  /*----------------------------------------------------------------*/
  FlushData ( & SuperVisService ) ;
  SuperVisService . CommandBuffer . CommandLineIndex = 0 ;
} /* SuperVFlushData */


static void SubDevFlushData ( void )
{
  /*----------------------------------------------------------------*/
  FlushData ( & SlaveDevService ) ;
  SlaveDevService . CommandBuffer . CommandLineIndex = 0 ;
} /* SubDevFlushData */


static uint8_t CompareDeviceID ( uint8_t * DeviceName ,
                                 uint8_t * Pattern )
{
  uint8_t Ch1 ;
  uint8_t Ch2 ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Ch1 = * DeviceName ++ ;
    Ch2 = * Pattern ++ ;
    if ( Ch1 != Ch2 )
    {
      return ( false ) ;
    } /* if */ ;
    if ( ( Ch1 == null ) || ( Ch2 == null ) )
      return ( true ) ;
  } /* for */ ;
} /* CompareDeviceID */


static void ClearBlock ( void *   Destination ,
                         uint16_t Size )
{
  uint16_t Loop ;
  uint8_t * Dst ;
  /*----------------------------------------------------------------*/
  Dst = ( uint8_t * ) Destination ;
  for ( Loop = 0 ; Loop < Size ; Loop ++ )
    * Dst ++ = null ;
} /* ClearBlock */


static void StrCopy ( uint8_t * Destination ,
                      uint8_t * Source )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = * Source ;
  * Destination = Ch ;
  while ( Ch )
  {
    Ch = * Source ++ ;
    * Destination ++ = Ch ;
  } /* while */ ;
} /* StrCopy */


static void ClearScreenForDevice ( uint8_t DeviceEntry )
{
  PtrSideServiceType SideData ;
  /*----------------------------------------------------------------*/
  SideData = & SlaveDevService ;
  SendTxt ( SideData , EEBlock . SubDeviceID [ DeviceEntry ] ) ;
  Send ( SideData , PartSeparator ) ;
  SendTxt ( SideData , EEBlock . MyDeviceID ) ;
  Send ( SideData , PartSeparator ) ;
  SendFlashTxt ( SideData , ( uint16_t ) ClearScreenCommand ) ;
  Send ( SideData , cr ) ;
//  SubDevFlushData ( ) ;
} /* ClearScreenForDevice  */


static uint8_t LocateParam ( uint8_t * ParamNamePattern )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < DisplParamArraySize ; Loop ++ )
  {
    if ( EEBlock . DisplParamDescr [ Loop ] [ 0 ] )
    {
      if ( CompareDeviceID ( EEBlock . DisplParamDescr [ Loop ] , ParamNamePattern ) )
      {
        return Loop ;
      } /* if */ ;
    } /* if */ ;
  } /* for */ ;
  return NoActive ;
} /* LocateParam */


static uint8_t LocateDevice ( uint8_t * DeviceNamePattern )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
  {
    if ( EEBlock . SubDeviceID [ Loop ] [ 0 ] )
    {
      if ( CompareDeviceID ( EEBlock . SubDeviceID [ Loop ] , DeviceNamePattern ) )
      {
        return Loop ;
      } /* if */ ;
    } /* if */ ;
  } /* for */ ;
  return NoActive ;
} /* LocateDevice */


static void ExecuteAssociation ( uint8_t ParamEntry ,
                                 uint8_t DeviceEntry )
{
  PtrSideServiceType SideData ;
  /*----------------------------------------------------------------*/
  SideData = & SlaveDevService ;
  SendTxt ( SideData , EEBlock . SubDeviceID [ DeviceEntry ] ) ;
  Send ( SideData , PartSeparator ) ;
  SendTxt ( SideData , EEBlock . MyDeviceID ) ;
  Send ( SideData , PartSeparator ) ;
  SendFlashTxt ( SideData , ( uint16_t ) DisplayAsciiCommand ) ;
  Send ( SideData , ParamSeparator ) ;
  SendTxt ( SideData , ParamValueArray [ ParamEntry ] . ParamValue ) ;
  Send ( SideData , cr ) ;
//  SubDevFlushData ( ) ;
} /* ExecuteAssociation */


static void DelAssociation ( uint8_t DeviceEntry )
{
  uint8_t Loop ;
  uint8_t IntLoop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < DisplParamArraySize ; Loop ++ )
  {
    for ( IntLoop = 0 ; IntLoop < ChannelsArraySize ; IntLoop ++ )
    {
      if ( ParamValueArray [ Loop ] . SubDeviceLink [ IntLoop ] == DeviceEntry )
      {
        ParamValueArray [ Loop ] . SubDeviceLink [ IntLoop ] = NoActive ;
      } /* if */ ;
    } /* for */ ;
  } /* for */ ;
} /* DelAssociation */


static void AddAssociation ( uint8_t ParamEntry ,
                             uint8_t DeviceEntry )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
  {
    if ( ParamValueArray [ ParamEntry ] . SubDeviceLink [ Loop ] == NoActive )
    {
      ParamValueArray [ ParamEntry ] . SubDeviceLink [ Loop ] = DeviceEntry ;
      break ;
    } /* if */ ;
  } /* for */ ;
} /* AddAssociation */


static void CopyCommandParamCont ( uint8_t *          SubDeviceID ,
                                   PtrSideServiceType SideData )
{
  uint8_t Loop ;
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  ClearBlock ( SubDeviceID , sizeof ( DeviceIDType ) ) ;
  for ( Loop = 0 ; Loop < DeviceIDSize ; Loop ++ )
  {
    Ch = SideData -> ParamText [ Loop ] ;
    if ( ! Ch )
      break ;
    * SubDeviceID ++ = Ch ;
  } /* for */ ;
} /* CopyCommandParamCont */


static void CodePassword ( void )
{
  uint8_t Loop ;
  uint8_t Element ;
  uint8_t BitCorrection ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Element = EEBlock . Password [ Loop ] ^ 0xFF ;
    BitCorrection = Element & 0x01 ;
    Element = Element >> 1 ;
    if ( BitCorrection )
      Element |= 0x80 ;
    EEBlock . Password [ Loop ] = Element ;
  } /* for */ ;
} /* CodePassword */


static void DecodePassword ( void )
{
  uint8_t Loop ;
  uint16_t Element ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Element = ( uint16_t ) EEBlock . Password [ Loop ] ;
    Element = Element << 1 ;
    if ( Element >= 0x100 )
      Element ++ ;
    Element &= 0xFF ;
    Element ^= 0xFF ;
    EEBlock . Password [ Loop ] = ( uint8_t ) Element ;
  } /* for */ ;
} /* DecodePassword */


static void ReadEEBlock ( void )
{
  /*----------------------------------------------------------------*/
  eeprom_read_block ( & EEBlock , ( void * ) EEBlockAddressOffset ,
                      sizeof ( EEPROMBlockDataRecT ) ) ;
  DecodePassword ( ) ;
} /* ReadEEBlock */


static void WriteEEBlock ( void )
{
  /*----------------------------------------------------------------*/
  LEDOn ( EEWriteLED ) ;
  CodePassword ( ) ;
  wdt_disable ( ) ;
  eeprom_write_block ( & EEBlock , ( void * ) EEBlockAddressOffset ,
                       sizeof ( EEPROMBlockDataRecT ) ) ;
  wdt_enable ( WDTO_1S ) ;
  LEDOff ( EEWriteLED ) ;
  DecodePassword ( ) ;
} /* WriteEEBlock */


static uint16_t ComputeCRC ( uint16_t  Length ,
                             uint8_t * Buffer )
{
  uint16_t Loop ;
  uint16_t ResultCRC ;
  uint16_t WorkCRC ;
  /*----------------------------------------------------------------*/
  ResultCRC = 0xFFFF ;
  for ( Loop = 0 ; Loop < Length ; Loop ++ )
  {
    WorkCRC = CRCTable [ ResultCRC & 0x0F ] ;
    ResultCRC = ( ResultCRC >> 4 ) & 0x0FFF ;
    ResultCRC ^= WorkCRC ^ CRCTable [ Buffer [ Loop ] & 0x0F ] ;
    WorkCRC = CRCTable [ ResultCRC & 0x0F ] ;
    ResultCRC = ( ResultCRC >> 4 ) & 0x0FFF ;
    ResultCRC ^= WorkCRC ^ CRCTable [ ( Buffer [ Loop ] >> 4 ) & 0x0F ] ;
  } /* for */ ;
  return ( ResultCRC ) ;
} /* ComputeCRC */


static void EEBlockOperation ( void )
{
  uint16_t CRCValue ;
  /*----------------------------------------------------------------*/
  ReadEEBlock ( ) ;
  if ( EEBlock . EEBlockMetric != StandardSygnature )
  {
    SetDefaultEEBlock ( ) ;
    WriteEEBlock ( ) ;
    return ;
  } /* if */ ;
  CRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) ,
                          ( uint8_t * ) ( & EEBlock ) ) ;
  if ( CRCValue != 0 )
  {
    SetDefaultEEBlock ( ) ;
    WriteEEBlock ( ) ;
  } /* if */ ;
} /* EEBlockOperation */


static void LinkParamValues ( void )
{
  uint8_t Loop ;
  uint8_t IntLoop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < DisplParamArraySize ; Loop ++ )
  {
    for ( IntLoop = 0 ; IntLoop < ChannelsArraySize ; IntLoop ++ )
    {
      ParamValueArray [ Loop ] . SubDeviceLink [ IntLoop ] = NoActive ;
    } /* for */ ;
    ClearBlock ( ParamValueArray [ Loop ] . ParamValue , ParamValueSize + 1 ) ;
  } /* for */ ;
} /* LinkParamValues */


static uint8_t CompareFlash ( uint8_t * Text ,
                              uint16_t  Pattern )
{
  uint8_t Ch1 ;
  uint8_t Ch2 ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Ch1 = * Text ++ ;
    Ch2 = pgm_read_byte_near ( Pattern ++ ) ;
    if ( Ch1 != Ch2 )
    {
      return ( false ) ;
    } /* if */ ;
    if ( Ch1 == 0 )
      return ( true ) ;
  } /* for */ ;
} /* CompareFlash */


static void SendTranslationCToN ( PtrSideServiceType SideData ,
                                  uint8_t            KeyCode ,
                                  uint8_t            AssociationTableSize ,
                                  uint16_t           AssociationTable )
{
  uint8_t Loop ;
  CodeAndNameAssociationRecT AssociationElement ;
  uint16_t MemAddress ;
  /*----------------------------------------------------------------*/
  MemAddress = AssociationTable ;
  for ( Loop = 0 ; Loop < AssociationTableSize ; Loop ++ )
  {
    AssociationElement . CodeElement = pgm_read_byte_near ( MemAddress ++ ) ;
    AssociationElement . NameElement = ( uint8_t * ) pgm_read_word_near ( MemAddress ) ;
    MemAddress += 2 ;
    if ( AssociationElement . CodeElement == KeyCode )
    {
      SendFlashTxt ( SideData , ( uint16_t ) AssociationElement . NameElement ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
} /* SendTranslationCToN */


static uint8_t TranslationNToC ( uint8_t * String ,
                                 uint8_t   AssociationTableSize ,
                                 uint16_t  AssociationTable )
{
  uint8_t Loop ;
  CodeAndNameAssociationRecT AssociationElement ;
  uint16_t MemAddress ;
  /*----------------------------------------------------------------*/
  MemAddress = AssociationTable ;
  for ( Loop = 0 ; Loop < AssociationTableSize ; Loop ++ )
  {
    AssociationElement . CodeElement = pgm_read_byte_near ( MemAddress ++ ) ;
    AssociationElement . NameElement = ( uint8_t * ) pgm_read_word_near ( MemAddress ) ;
    MemAddress += 2 ;
    if ( CompareFlash ( String , ( uint16_t ) AssociationElement . NameElement ) )
    {
      return ( AssociationElement . CodeElement ) ;
    } /* if */ ;
  } /* for */ ;
  return ( NoActive ) ;
} /* TranslationNToC */


static void SendKeyName ( PtrSideServiceType SideData ,
                          uint8_t            KeyCode )
{
  /*----------------------------------------------------------------*/
  SendTranslationCToN ( SideData , KeyCode , KeyCodeAndNameAssociationTabeSize ,
                        ( uint16_t ) KeyCodeNameAssociationTabe ) ;
} /* SendKeyName */


static uint8_t RouteToSubDev ( uint8_t * Destination )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
  {
    if ( EEBlock . SubDeviceID [ Loop ] [ 0 ] )
    {
      if ( CompareDeviceID ( Destination , EEBlock . SubDeviceID [ Loop ] ) )
        return true ;
    } /* if */ ;
  } /* for */ ;
  return false ;
} /* RouteToSubDev */


static uint8_t MeDestination ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  return ( CompareDeviceID ( SideData -> DestinationDeviceName , EEBlock . MyDeviceID ) ) ;
} /* MeDestination */


static uint8_t GeneralDestination ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  return ( CompareFlash ( SideData -> DestinationDeviceName , ( uint16_t ) GeneralPoolDeviceName ) ) ;
} /* GeneralDestination */


static void SendReplyForCommand ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  Send ( SideData , 'R' ) ;
  SendTxt ( SideData , SideData -> CommandText ) ;
  Send ( SideData , ParamSeparator ) ;
} /* SendReplyForCommand */


static void SendReplyDevicesID ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  SendTxt ( SideData , SideData -> SourceDeviceName ) ;
  Send ( SideData , PartSeparator ) ;
  SendTxt ( SideData , SideData -> DestinationDeviceName ) ;
  Send ( SideData , PartSeparator ) ;
} /* SendReplyDevicesID */


static void SendCommandReplyOK ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  SendReplyDevicesID ( SideData ) ;
  SendReplyForCommand ( SideData ) ;
  SendFlashTxt ( SideData , ( uint16_t ) OK ) ;
} /* SendCommandReplyOK */


static void SendCommandReplyFail ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  SendReplyDevicesID ( SideData ) ;
  SendReplyForCommand ( SideData ) ;
  SendFlashTxt ( SideData , ( uint16_t ) Fail ) ;
} /* SendCommandReplyFail */


static void ConnectCommandExecute ( PtrSideServiceType SideData )
{
  uint8_t Ch ;
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Ch = EEBlock . Password [ Loop ] ;
    if ( SideData -> ParamText [ Loop ] != Ch )
    {
      SendCommandReplyFail ( SideData ) ;
      return ;
    } /* if */ ;
    if ( Ch == null )
      break ;
  } /* for */ ;
  ConnectContext = true ;
  SendCommandReplyOK ( SideData ) ;
  LEDOn ( ConnectLED ) ;
} /* ConnectCommandExecute */


static void DisConnectCommandExecute ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  ConnectContext = false ;
  SendCommandReplyOK ( SideData ) ;
  LEDOff ( ConnectLED ) ;
} /* DisConnectCommandExecute */


static uint8_t ExecuteAssociatedCommand ( PtrSideServiceType SideData ,
                                          uint8_t *          String ,
                                          uint8_t            AssociationTableSize ,
                                          uint16_t           AssociationTable )
{
  uint8_t Loop ;
  CommandNameAssociationRecT AssociationElement ;
  uint16_t MemAddress ;
  /*----------------------------------------------------------------*/
  MemAddress = AssociationTable ;
  for ( Loop = 0 ; Loop < AssociationTableSize ; Loop ++ )
  {
    AssociationElement . CommandService = ( CommandExecuteServiceProcT ) pgm_read_word_near ( MemAddress ) ;
    MemAddress += 2 ;
    AssociationElement . CommandName = ( uint8_t * ) pgm_read_word_near ( MemAddress ) ;
    MemAddress += 2 ;
    if ( CompareFlash ( String , ( uint16_t ) AssociationElement . CommandName ) )
    {
      AssociationElement . CommandService ( SideData ) ;
      return ( true ) ;
    } /* if */ ;
  } /* for */ ;
  return ( false ) ;
} /* ExecuteAssociatedCommand */


static void HelloCommandExecute ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  SendReplyDevicesID ( SideData ) ;
  SendReplyForCommand ( SideData ) ;
  SendFlashTxt ( SideData , ( uint16_t ) DeviceName ) ;
  SuperVFlushData ( ) ;
} /* HelloCommandExecute */


static void ListSubDevCommandExecute ( PtrSideServiceType SideData )
{
  uint8_t Loop ;
  uint8_t AnyDataPresent ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  AnyDataPresent = false ;
  for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
  {
    if ( EEBlock . SubDeviceID [ Loop ] [ 0 ] )
    {
      AnyDataPresent = true ;
      SendReplyDevicesID ( SideData ) ;
      SendReplyForCommand ( SideData ) ;
      SendTxt ( SideData , EEBlock . SubDeviceID [ Loop ] ) ;
      Send ( SideData , cr ) ;
      SuperVFlushData ( ) ;
    } /* if */ ;
  } /* for */ ;
  if ( ! AnyDataPresent )
  {
    SendReplyDevicesID ( SideData ) ;
    SendReplyForCommand ( SideData ) ;
    SendFlashTxt ( SideData , ( uint16_t ) EmptyData ) ;
    SuperVFlushData ( ) ;
  } /* if */ ;
} /* ListSubDevCommandExecute */


static void GetVersionCommandExecute ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  SendReplyDevicesID ( SideData ) ;
  SendReplyForCommand ( SideData ) ;
  SendFlashTxt ( SideData , ( uint16_t ) HardwareSoftwareVersion ) ;
  SuperVFlushData ( ) ;
} /* GetVersionCommandExecute */


static void GetIDCommandExecute ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  SendReplyDevicesID ( SideData ) ;
  SendReplyForCommand ( SideData ) ;
  SendTxt ( SideData , EEBlock . MyDeviceID ) ;
  Send ( SideData , cr ) ;
  SuperVFlushData ( ) ;
} /* GetIDCommandExecute */


static void ReinitCommandExecute ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  EEBlockOperation ( ) ;
  SendCommandReplyOK ( SideData ) ;
  SuperVFlushData ( ) ;
} /* ReinitCommandExecute */


static void ResetEECommandExecute ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  SetDefaultEEBlock ( ) ;
  SendCommandReplyOK ( SideData ) ;
  SuperVFlushData ( ) ;
} /* ResetEECommandExecute */


static void SaveEECommandExecute ( PtrSideServiceType SideData )
{
  uint16_t CRCValue ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  LEDOn ( EEWriteLED ) ;
  EEBlock . CRCValue = 0 ;
  CRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) -
                          sizeof ( uint16_t ) ,
                          ( uint8_t * ) ( & EEBlock ) ) ;
  EEBlock . CRCValue = CRCValue ;
  WriteEEBlock ( ) ;
  SendCommandReplyOK ( SideData ) ;
  SuperVFlushData ( ) ;
  LEDOff ( EEWriteLED ) ;
} /* SaveEECommandExecute */


static uint8_t DecodeKeyName ( uint8_t * String )
{
  /*----------------------------------------------------------------*/
  return ( TranslationNToC ( String , KeyCodeAndNameAssociationTabeSize ,
                             ( uint16_t ) KeyCodeNameAssociationTabe ) ) ;
} /* DecodeKeyName */


static uint8_t AnyParamPresent ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < DisplParamArraySize ; Loop ++ )
  {
    if ( EEBlock . DisplParamDescr [ Loop ] [ 0 ] )
      return true ;
  } /* for */ ;
  return false ;
} /* AnyParamPresent */


static void ListParamCommandExecute ( PtrSideServiceType SideData )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( AnyParamPresent ( ) )
  {
    for ( Loop = 0 ; Loop < DisplParamArraySize ; Loop ++ )
    {
      if ( EEBlock . DisplParamDescr [ Loop ] [ 0 ] )
      {
        SendReplyDevicesID ( SideData ) ;
        SendReplyForCommand ( SideData ) ;
        SendTxt ( SideData , EEBlock . DisplParamDescr [ Loop ] ) ;
        Send ( SideData , cr ) ;
        SuperVFlushData ( ) ;
      } /* if */ ;
    } /* for */ ;
  } /* if ... */
  else
  {
    SendReplyDevicesID ( SideData ) ;
    SendReplyForCommand ( SideData ) ;
    SendFlashTxt ( SideData , ( uint16_t ) EmptyData ) ;
    SuperVFlushData ( ) ;
  } /* if ... else */ ;
} /* ListParamCommandExecute */


static void DisplayParamCommandExecute ( PtrSideServiceType SideData )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( AnyParamPresent ( ) )
  {
    for ( Loop = 0 ; Loop < DisplParamArraySize ; Loop ++ )
    {
      if ( EEBlock . DisplParamDescr [ Loop ] [ 0 ] )
      {
        SendReplyDevicesID ( SideData ) ;
        SendReplyForCommand ( SideData ) ;
        SendTxt ( SideData , EEBlock . DisplParamDescr [ Loop ] ) ;
        Send ( SideData , ParamSeparator ) ;
        if ( ParamValueArray [ Loop ] . ParamValue [ 0 ] )
        {
          SendTxt ( SideData , ParamValueArray [ Loop ] . ParamValue ) ;
          Send ( SideData , cr ) ;
        } /* if ... */
        else
          SendFlashTxt ( SideData , ( uint16_t ) EmptyData ) ;
        SuperVFlushData ( ) ;
      } /* if */ ;
    } /* for */ ;
  } /* if ... */
  else
  {
    SendReplyDevicesID ( SideData ) ;
    SendReplyForCommand ( SideData ) ;
    SendFlashTxt ( SideData , ( uint16_t ) EmptyData ) ;
    SuperVFlushData ( ) ;
  } /* if ... else */ ;
} /* DisplayParamCommandExecute */


static void ParamValueCommandExecute ( PtrSideServiceType SideData )
{
  uint8_t Loop ;
  uint8_t IndexInParamArray ;
  uint8_t Ch ;
  uint8_t ParamEntry ;
  ParamNameType TmpParamName ;
  /*----------------------------------------------------------------*/
  if ( ! SideData -> ParamText [ 0 ] )
  {
    SendCommandReplyFail ( SideData ) ;
    SuperVFlushData ( ) ;
    return ;
  } /* if */ ;
  ClearBlock ( TmpParamName , ParamNameSize + 1 ) ;
  IndexInParamArray = 0 ;
  for ( Loop = 0 ; Loop <= ParamNameSize ; Loop ++ )
  {
    Ch = SideData -> ParamText [ IndexInParamArray ] ;
    IndexInParamArray ++ ;
    if ( ! Ch )
    {
      SendCommandReplyFail ( SideData ) ;
      SuperVFlushData ( ) ;
      return ;
    } /* if */ ;
    if ( Ch == ParamSeparator )
      break ;
    if ( Loop < ParamNameSize )
      TmpParamName [ Loop ] = Ch ;
  } /* for */ ;
  ParamEntry = LocateParam ( TmpParamName ) ;
  if ( ParamEntry != NoActive )
  {
    for ( Loop = 0 ; Loop <= ParamValueSize ; Loop ++ )
      ParamValueArray [ ParamEntry ] . ParamValue [ Loop ] = 0 ;
    for ( Loop = 0 ; Loop < ParamValueSize ; Loop ++ )
    {
      Ch = SideData -> ParamText [ IndexInParamArray ] ;
      IndexInParamArray ++ ;
      if ( ( ! Ch ) || ( Ch == '.' ) )
        break ;
      ParamValueArray [ ParamEntry ] . ParamValue [ Loop ] = Ch ;
    } /* for */ ;
    for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
    {
      if ( ParamValueArray [ ParamEntry ] . SubDeviceLink [ Loop ] != NoActive )
      {
        ExecuteAssociation ( ParamEntry , ParamValueArray [ ParamEntry ] . SubDeviceLink [ Loop ] ) ;
      } /* if */ ;
    } /* for */ ;
    SendCommandReplyOK ( SideData ) ;
    SuperVFlushData ( ) ;
    return ;
  } /* if */ ;
  SendCommandReplyFail ( SideData ) ;
  SuperVFlushData ( ) ;
} /* ParamValueCommandExecute */


static void AddChannelCommandExecute ( PtrSideServiceType SideData )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SideData -> ParamText [ 0 ] )
  {
    SendCommandReplyFail ( SideData ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
  {
    if ( ! EEBlock . SubDeviceID [ Loop ] [ 0 ] )
    {
      CopyCommandParamCont ( EEBlock . SubDeviceID [ Loop ] , SideData ) ;
      SendCommandReplyOK ( SideData ) ;
      SuperVFlushData ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  SendCommandReplyFail ( SideData ) ;
  SuperVFlushData ( ) ;
} /* AddChannelCommandExecute */


static void DelChannelCommandExecute ( PtrSideServiceType SideData )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SideData -> ParamText [ 0 ] )
  {
    SendCommandReplyFail ( SideData ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
  {
    if ( CompareDeviceID ( EEBlock . SubDeviceID [ Loop ] , SideData -> ParamText ) )
    {
      ClearBlock ( EEBlock . SubDeviceID [ Loop ] , sizeof ( DeviceIDType ) ) ;
      DelAssociation ( Loop ) ;
      SendCommandReplyOK ( SideData ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  SendCommandReplyFail ( SideData ) ;
  SuperVFlushData ( ) ;
} /* DelChannelCommandExecute */


static void SetDeviceNameCommandExecute ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SideData -> ParamText [ 0 ] )
  {
    SendCommandReplyFail ( SideData ) ;
    return ;
  } /* if */ ;
  CopyCommandParamCont ( EEBlock . MyDeviceID , SideData ) ;
  SendCommandReplyOK ( SideData ) ;
  SuperVFlushData ( ) ;
} /* SetDeviceNameCommandExecute */


static void StatusDispModeCommandExecute ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  LEDModeSwitch = StatusDispMode ;
  EEBlock . DisplayMode = StatusDispMode ;
  SendCommandReplyOK ( SideData ) ;
  SuperVFlushData ( ) ;
} /* StatusDispModeCommandExecute */


static void FunctionDispModeCommandExecute ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  LEDModeSwitch = FunctionDispMode ;
  EEBlock . DisplayMode = FunctionDispMode ;
  SendCommandReplyOK ( SideData ) ;
  SuperVFlushData ( ) ;
} /* FunctionDispModeCommandExecute */


static void AddParamCommandExecute ( PtrSideServiceType SideData )
{
  uint8_t Loop ;
  uint8_t ParamEntry ;
  ParamNameType TmpParamName ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SideData -> ParamText [ 0 ] )
  {
    SendCommandReplyFail ( SideData ) ;
    SuperVFlushData ( ) ;
    return ;
  } /* if */ ;
  SideData -> ParamText [ ParamNameSize ] = 0 ;
  ClearBlock ( TmpParamName , ParamNameSize + 1 ) ;
  StrCopy ( TmpParamName , SideData -> ParamText ) ;
  ParamEntry = LocateParam ( TmpParamName ) ;
  if ( ParamEntry != NoActive )
  {
    SendCommandReplyFail ( SideData ) ;
    SuperVFlushData ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < DisplParamArraySize ; Loop ++ )
  {
    if ( ! EEBlock . DisplParamDescr [ Loop ] [ 0 ] )
    {
      StrCopy ( EEBlock . DisplParamDescr [ Loop ] , TmpParamName ) ;
      SendCommandReplyOK ( SideData ) ;
      SuperVFlushData ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  SendCommandReplyFail ( SideData ) ;
  SuperVFlushData ( ) ;
} /* AddParamCommandExecute */


static void DelParamCommandExecute ( PtrSideServiceType SideData )
{
  uint8_t Loop ;
  uint8_t DeviceEntry ;
  uint8_t ParamEntry ;
  ParamNameType TmpParamName ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SideData -> ParamText [ 0 ] )
  {
    SendCommandReplyFail ( SideData ) ;
    SuperVFlushData ( ) ;
    return ;
  } /* if */ ;
  SideData -> ParamText [ ParamNameSize ] = 0 ;
  ClearBlock ( TmpParamName , ParamNameSize + 1 ) ;
  StrCopy ( TmpParamName , SideData -> ParamText ) ;
  ParamEntry = LocateParam ( TmpParamName ) ;
  if ( ParamEntry == NoActive )
  {
    SendCommandReplyFail ( SideData ) ;
    SuperVFlushData ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
  {
    DeviceEntry = ParamValueArray [ ParamEntry ] . SubDeviceLink [ Loop ] ;
    if ( DeviceEntry != NoActive )
    {
      ClearScreenForDevice ( DeviceEntry ) ;
    } /* if */ ;
  } /* for */ ;
  ClearBlock ( ParamValueArray [ ParamEntry ] . ParamValue , ParamValueSize + 1 ) ;
  for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
  {
    ParamValueArray [ ParamEntry ] . SubDeviceLink [ Loop ] = NoActive ;
  } /* for */ ;
  SendCommandReplyOK ( SideData ) ;
  SuperVFlushData ( ) ;
} /* DelParamCommandExecute */


static void SetStartAssocCommandExecute ( PtrSideServiceType SideData )
{
  uint8_t Loop ;
  uint8_t KeyCode ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SideData -> ParamText [ 0 ] )
  {
    SendCommandReplyFail ( SideData ) ;
    SuperVFlushData ( ) ;
    return ;
  } /* if */ ;
  KeyCode = DecodeKeyName ( SideData -> ParamText ) ;
  if ( KeyCode == NoActive )
  {
    SendCommandReplyFail ( SideData ) ;
    SuperVFlushData ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < KeyCommTableSize ; Loop ++ )
  {
    if ( EEBlock . KeyCommTable [ Loop ] . KeyCode == KeyCode )
    {
      EEBlock . StartKeyCode = KeyCode ;
      SendCommandReplyOK ( SideData ) ;
      SuperVFlushData ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  SendCommandReplyFail ( SideData ) ;
  SuperVFlushData ( ) ;
} /* SetStartAssocCommandExecute */


static void ResetStartAssocCommandExecute ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  EEBlock . StartKeyCode = NoActive ;
  SendCommandReplyOK ( SideData ) ;
  SuperVFlushData ( ) ;
} /* ResetStartAssocCommandExecute */


static void GetStartAssocCommandExecute ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  SendReplyDevicesID ( SideData ) ;
  SendReplyForCommand ( SideData ) ;
  if ( EEBlock . StartKeyCode == NoActive )
  {
    SendFlashTxt ( SideData , ( uint16_t ) EmptyData ) ;
  } /* if ... */
  else
  {
    SendKeyName ( SideData , EEBlock . StartKeyCode ) ;
    Send ( SideData , cr ) ;
  } /* if ... else */ ;
  SuperVFlushData ( ) ;
} /* GetStartAssocCommandExecute */


static void DisplayTableElement ( PtrSideServiceType  SideData ,
                                  PtrKeyCommDescrRecT CommDescr )
{
  /*----------------------------------------------------------------*/
  SendReplyDevicesID ( SideData ) ;
  SendReplyForCommand ( SideData ) ;
  SendKeyName ( SideData , CommDescr -> KeyCode ) ;
  Send ( SideData , ParamSeparator ) ;
  SendTxt ( SideData , CommDescr -> ParamName ) ;
  Send ( SideData , ParamSeparator ) ;
  SendTxt ( SideData , CommDescr -> DestDevice ) ;
  Send ( SideData , cr ) ;
  SuperVFlushData ( ) ;
} /* DisplayTableElement */


static void ListKeyTableCommandExecute ( PtrSideServiceType SideData )
{
  uint8_t Loop ;
  uint8_t AnyPresent ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  AnyPresent = false ;
  for ( Loop = 0 ; Loop < KeyCommTableSize ; Loop ++ )
  {
    if ( EEBlock . KeyCommTable [ Loop ] . KeyCode != NoActive )
    {
      DisplayTableElement ( SideData , & EEBlock . KeyCommTable [ Loop ] ) ;
      AnyPresent = true ;
    } /* if */ ;
  } /* for */ ;
  if ( ! AnyPresent )
  {
    SendReplyDevicesID ( SideData ) ;
    SendReplyForCommand ( SideData ) ;
    SendFlashTxt ( SideData , ( uint16_t ) EmptyData ) ;
    SuperVFlushData ( ) ;
  } /* if */ ;
} /* ListKeyTableCommandExecute */


static uint8_t Unpack_Key_Param_Device ( PtrKeyCommDescrRecT Element ,
                                         PtrSideServiceType  SideData )
{
  DeviceIDType KeyName ;
  uint8_t Loop ;
  uint8_t IndexInParamArray ;
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( ! SideData -> ParamText [ 0 ] )
  {
    return false ;
  } /* if */ ;
  ClearBlock ( Element , sizeof ( KeyCommDescrRecT ) ) ;
  ClearBlock ( KeyName , DeviceIDSize + 1 ) ;
  IndexInParamArray = 0 ;
  for ( Loop = 0 ; ; Loop ++ )
  {
    Ch = SideData -> ParamText [ IndexInParamArray ] ;
    IndexInParamArray ++ ;
    if ( ! Ch )
    {
      return false ;
    } /* if */ ;
    if ( Ch == ParamSeparator )
      break ;
    if ( Loop < DeviceIDSize )
      KeyName [ Loop ] = Ch ;
  } /* for */ ;
  for ( Loop = 0 ; ; Loop ++ )
  {
    Ch = SideData -> ParamText [ IndexInParamArray ] ;
    IndexInParamArray ++ ;
    if ( ! Ch )
    {
      return false ;
    } /* if */ ;
    if ( Ch == ParamSeparator )
      break ;
    if ( Loop < ParamNameSize )
      Element -> ParamName [ Loop ] = Ch ;
  } /* for */ ;
  for ( Loop = 0 ; ; Loop ++ )
  {
    Ch = SideData -> ParamText [ IndexInParamArray ] ;
    IndexInParamArray ++ ;
    if ( ! Ch )
    {
      break ;
    } /* if */ ;
    if ( Loop < DeviceIDSize )
      Element -> DestDevice [ Loop ] = Ch ;
  } /* for */ ;
  Element -> KeyCode = DecodeKeyName ( KeyName ) ;
  if ( Element -> KeyCode == NoActive )
  {
    return false ;
  } /* if */ ;
  return true ;
} /* Unpack_Key_Param_Device */


static void AddKeyTableCommandExecute ( PtrSideServiceType SideData )
{
  KeyCommDescrRecT AddElement ;
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! Unpack_Key_Param_Device ( & AddElement , SideData ) )
  {
    SendCommandReplyFail ( SideData ) ;
    SuperVFlushData ( ) ;
    return ;
  } /* if */ ;
  if ( LocateParam ( AddElement . ParamName ) == NoActive )
  {
    SendCommandReplyFail ( SideData ) ;
    SuperVFlushData ( ) ;
    return ;
  } /* if */ ;
  if ( LocateDevice ( AddElement . DestDevice ) == NoActive )
  {
    SendCommandReplyFail ( SideData ) ;
    SuperVFlushData ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < KeyCommTableSize ;Loop ++ )
  {
    if ( EEBlock . KeyCommTable [ Loop ] . KeyCode == NoActive )
    {
      EEBlock . KeyCommTable [ Loop ] = AddElement ;
      SendCommandReplyOK ( SideData ) ;
      SuperVFlushData ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  SendCommandReplyFail ( SideData ) ;
  SuperVFlushData ( ) ;
} /* AddKeyTableCommandExecute */


static void DelKeyTableCommandExecute ( PtrSideServiceType SideData )
{
  KeyCommDescrRecT DelElement ;
  uint8_t Loop ;
  uint8_t DeviceEntry ;
  uint8_t ParamEntry ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! Unpack_Key_Param_Device ( & DelElement , SideData ) )
  {
    SendCommandReplyFail ( SideData ) ;
    SuperVFlushData ( ) ;
    return ;
  } /* if */ ;
  DeviceEntry = LocateDevice ( DelElement . DestDevice ) ;
  ParamEntry = LocateParam ( DelElement . ParamName ) ;
  if ( ( DeviceEntry == NoActive ) || ( ParamEntry == NoActive ) )
  {
    SendCommandReplyFail ( SideData ) ;
    SuperVFlushData ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < KeyCommTableSize ; Loop ++ )
  {
    if ( EEBlock . KeyCommTable [ Loop ] . KeyCode == DelElement . KeyCode )
    {
      if ( ! CompareDeviceID ( EEBlock . KeyCommTable [ Loop ] . ParamName , DelElement . ParamName ) )
        continue ;
      if ( ! CompareDeviceID ( EEBlock . KeyCommTable [ Loop ] . DestDevice , DelElement . DestDevice ) )
        continue ;
      EEBlock . KeyCommTable [ Loop ] . KeyCode = NoActive ;
      ClearBlock ( EEBlock . KeyCommTable [ Loop ] . DestDevice , DeviceIDSize + 1 ) ;
      ClearBlock ( EEBlock . KeyCommTable [ Loop ] . ParamName , ParamNameSize + 1 ) ;
      SendCommandReplyOK ( SideData ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  SendCommandReplyFail ( SideData ) ;
  SuperVFlushData ( ) ;
} /* DelKeyTableCommandExecute */


static void NewPasswordCommandExecute ( PtrSideServiceType SideData )
{
  uint8_t Ch ;
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Ch = SideData -> ParamText [ Loop ] ;
    EEBlock . Password [ Loop ] = Ch ;
    if ( Ch == null )
      break ;
  } /* for */ ;
  SendCommandReplyOK ( SideData ) ;
  SuperVFlushData ( ) ;
} /* NewPasswordCommandExecute */


static void ExecuteKeyCommand ( uint8_t KeyCode ) ;

static void RunKeyCommandExecute ( PtrSideServiceType SideData )
{
  uint8_t KeyCode ;
  /*----------------------------------------------------------------*/
  KeyCode = DecodeKeyName ( SideData -> ParamText ) ;
  if ( KeyCode == 0x0FF )
  {
    SendCommandReplyFail ( SideData ) ;
    return ;
  } /* if */ ;
  ExecuteKeyCommand ( KeyCode ) ;
} /* RunKeyCommandExecute */


static void SuperVisExecuteCommand ( void )
{
  uint8_t FrameToMe ;
  /*----------------------------------------------------------------*/
  FrameToMe = MeDestination ( & SuperVisService ) ;
  if ( ! FrameToMe )
  {
    FrameToMe = GeneralDestination ( & SuperVisService ) ;
    if ( FrameToMe )
      StrCopy ( SuperVisService . DestinationDeviceName ,
                EEBlock . MyDeviceID ) ;
  } /* if */ ;
  if ( FrameToMe )
  {
    if ( ! ExecuteAssociatedCommand ( & SuperVisService , SuperVisService . CommandText ,
                                      SvCommandNameAssociationTabeSize ,
                                      ( uint16_t ) SvCommandNameAssociationTabe ) )
    {
    } /* if */ ;
    SuperVFlushData ( ) ;
  } /* if ... */
  else
  {
    if ( RouteToSubDev ( SuperVisService . DestinationDeviceName ) )
    {
      SendTxt ( & SlaveDevService , SuperVisService . DestinationDeviceName ) ;
      Send ( & SlaveDevService , PartSeparator ) ;
      SendTxt ( & SlaveDevService , SuperVisService . SourceDeviceName ) ;
      Send ( & SlaveDevService , PartSeparator ) ;
      SendTxt ( & SlaveDevService , SuperVisService . CommandText ) ;
      if ( SuperVisService . ParamText [ 0 ] )
      {
        Send ( & SlaveDevService , ParamSeparator ) ;
        SendTxt ( & SlaveDevService , SuperVisService . ParamText ) ;
      } /* if */ ;
      Send ( & SlaveDevService , cr ) ;
    } /* if ... */
    else
    {
//               pakiet do kuba
    } /* if ... else */ ;
  } /* if ... else */ ;
} /* SuperVisExecuteCommand */


static void SubDevExecuteCommand ( void )
{
  uint8_t FrameToMe ;
  /*----------------------------------------------------------------*/
  FrameToMe = MeDestination ( & SlaveDevService ) ;
  if ( ! FrameToMe )
  {
    FrameToMe = GeneralDestination ( & SlaveDevService ) ;
    if ( FrameToMe )
      StrCopy ( SlaveDevService . DestinationDeviceName ,
                EEBlock . MyDeviceID ) ;
  } /* if */ ;
  if ( FrameToMe )
  {
//    ExecuteAssociatedCommand ( & SubDevService , SubDevService . CommandText ,
//                               SubDevCommandNameAssociationTabeSize ,
//                               ( uint16_t ) SubDevCommandNameAssociationTabe ) ;
  } /* if ... */
  else
  {
  } /* if ... else */ ;
} /* SubDevExecuteCommand */


static void ResetOperation ( PtrSideServiceType SideData ) ;

static void DestinationIDPutCh ( PtrSideServiceType SideData ,
                                 uint8_t            Ch )
{
  /*----------------------------------------------------------------*/
  if ( SideData -> DestDeviceNameIndex < DeviceIDSize )
  {
    SideData -> DestinationDeviceName [ SideData -> DestDeviceNameIndex ++ ] = Ch ;
  } /* if */ ;
} /* DestinationIDPutCh */


static void SourceIDPutCh ( PtrSideServiceType SideData ,
                            uint8_t            Ch )
{
  /*----------------------------------------------------------------*/
  if ( SideData -> SrcDeviceNameIndex < DeviceIDSize )
  {
    SideData -> SourceDeviceName [ SideData -> SrcDeviceNameIndex ++ ] = Ch ;
  } /* if */ ;
} /* SourceIDPutCh */


static void CommandIDPutCh ( PtrSideServiceType SideData ,
                             uint8_t            Ch )
{
  /*----------------------------------------------------------------*/
  if ( SideData -> CommandTextIndex < CommandBufferSize )
  {
    SideData -> CommandText [ SideData -> CommandTextIndex ++ ] = Ch ;
  } /* if */ ;
} /* CommandIDPutCh */


static void ParamPutCh ( PtrSideServiceType SideData ,
                         uint8_t            Ch )
{
  /*----------------------------------------------------------------*/
  if ( SideData -> ParamTextIndex < ParamBufferSize )
  {
    SideData -> ParamText [ SideData -> ParamTextIndex ++ ] = Ch ;
  } /* if */ ;
} /* ParamPutCh */


static void SelectParam ( PtrSideServiceType SideData )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SideData -> SerialDataPresent ( ) )
  {
    Ch = SideData -> GetSerial ( ) ;
    if ( Ch == cr )
    {
      ParamPutCh ( SideData , null ) ;
      SideData -> ExecuteCommand ( ) ;
      ResetOperation ( SideData ) ;
      return ;
    } /* if */;
    if ( Ch < Space )
    {
      ResetOperation ( SideData ) ;
      return ;
    } /* if */ ;
    ParamPutCh ( SideData , Ch ) ;
  } /* if */ ;
} /* SelectParam */


static void SelectCommand ( PtrSideServiceType SideData )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SideData -> SerialDataPresent ( ) )
  {
    Ch = SideData -> GetSerial ( ) ;
    if ( Ch == cr )
    {
      CommandIDPutCh ( SideData , null ) ;
      SideData -> ExecuteCommand ( ) ;
      ResetOperation ( SideData ) ;
      return ;
    } /* if */;
    if ( Ch < Space )
    {
      ResetOperation ( SideData ) ;
      return ;
    } /* if */ ;
    if ( Ch == ParamSeparator )
    {
      CommandIDPutCh ( SideData , null ) ;
      SideData -> SerialService = SelectParam ;
      return ;
    } /* if */;
    CommandIDPutCh ( SideData , Ch ) ;
  } /* if */ ;
} /* SelectCommand */


static void SelectSourceService ( PtrSideServiceType SideData )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SideData -> SerialDataPresent ( ) )
  {
    Ch = SideData -> GetSerial ( ) ;
    if ( Ch < Space )
    {
      ResetOperation ( SideData ) ;
      return ;
    } /* if */ ;
    if ( Ch == PartSeparator )
    {
      SourceIDPutCh ( SideData , null ) ;
      SideData -> SerialService = SelectCommand ;
      return ;
    } /* if */;
    SourceIDPutCh ( SideData , Ch ) ;
  } /* if */ ;
} /* SelectSourceService */


static void SelectDestOperation ( PtrSideServiceType SideData )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SideData -> SerialDataPresent ( ) )
  {
    Ch = SideData -> GetSerial ( ) ;
    if ( Ch < Space )
    {
      ResetOperation ( SideData ) ;
      return ;
    } /* if */ ;
    if ( Ch == PartSeparator )
    {
      DestinationIDPutCh ( SideData , null ) ;
      SideData -> SerialService = SelectSourceService ;
      return ;
    } /* if */;
    DestinationIDPutCh ( SideData , Ch ) ;
  } /* if */ ;
} /* SelectDestOperation */


static void ClearAllScreens ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
  {
    if ( EEBlock . SubDeviceID [ Loop ] [ 0 ] )
    {
      ClearScreenForDevice ( Loop ) ;
    } /* if */ ;
  } /* for */ ;
} /* ClearAllScreens */


static void SelectNewFunction ( uint8_t NewFunction )
{
  /*----------------------------------------------------------------*/
  ClearFunctionIndicator ( CurrentFunctionSelected ) ;
  CurrentFunctionSelected = NewFunction ;
  SetFunctionIndicator ( CurrentFunctionSelected ) ;
} /* SelectNewFunction */


static uint8_t LocateKeyCommand ( uint8_t KeyCode )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < KeyCommTableSize ; Loop ++ )
  {
    if ( EEBlock . KeyCommTable [ Loop ] . KeyCode == KeyCode )
    {
      return true ;
    } /* if */ ;
  } /* for */ ;
  return false ;
} /* LocateKeyCommand */


static void CreateNewAssociation ( uint8_t KeyCode )
{
  uint8_t Loop ;
  uint8_t ParamEntry ;
  uint8_t DeviceEntry ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < KeyCommTableSize ; Loop ++ )
  {
    if ( EEBlock . KeyCommTable [ Loop ] . KeyCode == KeyCode )
    {
      ParamEntry = LocateParam ( EEBlock . KeyCommTable [ Loop ] . ParamName ) ;
      if ( ParamEntry == NoActive )
        continue ;
      DeviceEntry = LocateDevice ( EEBlock . KeyCommTable [ Loop ] . DestDevice ) ;
      if ( DeviceEntry == NoActive )
        continue ;
      ClearScreenForDevice ( DeviceEntry ) ;
      DelAssociation ( DeviceEntry ) ;
      AddAssociation ( ParamEntry , DeviceEntry ) ;
      ExecuteAssociation ( ParamEntry , DeviceEntry ) ;
    } /* if */ ;
  } /* for */ ;
} /* CreateNewAssociation */


static void ExecuteKeyCommand ( uint8_t KeyCode )
{
  /*----------------------------------------------------------------*/
  switch ( KeyCode )
  {
    case StandKey0Code    :
      if ( LocateKeyCommand ( StandKey0Code ) )
      {
        SelectNewFunction ( Function0Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case StandKey1Code    :
      if ( LocateKeyCommand ( StandKey1Code ) )
      {
        SelectNewFunction ( Function1Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case StandKey2Code    :
      if ( LocateKeyCommand ( StandKey2Code ) )
      {
        SelectNewFunction ( Function2Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case StandKey3Code    :
      if ( LocateKeyCommand ( StandKey3Code ) )
      {
        SelectNewFunction ( Function3Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case StandKey4Code    :
      if ( LocateKeyCommand ( StandKey4Code ) )
      {
        SelectNewFunction ( Function4Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case StandKey5Code    :
      if ( LocateKeyCommand ( StandKey5Code ) )
      {
        SelectNewFunction ( Function5Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case StandKey6Code    :
      if ( LocateKeyCommand ( StandKey6Code ) )
      {
        SelectNewFunction ( Function6Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case StandKey7Code    :
      if ( LocateKeyCommand ( StandKey7Code ) )
      {
        SelectNewFunction ( Function7Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case StandKey01Code   :
      LEDModeSwitch = FunctionDispMode ;
      break ;
    case StandKey02Code   :
      break ;
    case StandKey03Code   :
      break ;
    case StandKey04Code   :
      break ;
    case StandKey05Code   :
      break ;
    case StandKey06Code   :
      break ;
    case StandKey07Code   :
      break ;
    case StandKey12Code   :
      LEDModeSwitch = StatusDispMode ;
      break ;
    case StandKey13Code   :
      break ;
    case StandKey14Code   :
      break ;
    case StandKey15Code   :
      break ;
    case StandKey16Code   :
      break ;
    case StandKey17Code   :
      break ;
    case StandKey23Code   :
      break ;
    case StandKey24Code   :
      break ;
    case StandKey25Code   :
      break ;
    case StandKey26Code   :
      break ;
    case StandKey27Code   :
      break ;
    case StandKey34Code   :
      break ;
    case StandKey35Code   :
      break ;
    case StandKey36Code   :
      break ;
    case StandKey37Code   :
      break ;
    case StandKey45Code   :
      break ;
    case StandKey46Code   :
      break ;
    case StandKey47Code   :
      break ;
    case StandKey56Code   :
      break ;
    case StandKey57Code   :
      break ;
    case StandKey67Code   :
      break ;
    case LongKey0Code     :
      if ( LocateKeyCommand ( LongKey0Code ) )
      {
        SelectNewFunction ( Function0Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case LongKey1Code     :
      if ( LocateKeyCommand ( LongKey1Code ) )
      {
        SelectNewFunction ( Function1Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case LongKey2Code     :
      if ( LocateKeyCommand ( LongKey2Code ) )
      {
        SelectNewFunction ( Function2Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case LongKey3Code     :
      if ( LocateKeyCommand ( LongKey3Code ) )
      {
        SelectNewFunction ( Function3Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case LongKey4Code     :
      if ( LocateKeyCommand ( LongKey4Code ) )
      {
        SelectNewFunction ( Function4Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case LongKey5Code     :
      if ( LocateKeyCommand ( LongKey5Code ) )
      {
        SelectNewFunction ( Function5Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case LongKey6Code     :
      if ( LocateKeyCommand ( LongKey6Code ) )
      {
        SelectNewFunction ( Function6Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case LongKey7Code     :
      if ( LocateKeyCommand ( LongKey7Code ) )
      {
        SelectNewFunction ( Function7Selected ) ;
        CreateNewAssociation ( KeyCode ) ;
      } /* if */ ;
      break ;
    case LongKey01Code    :
      break ;
    case LongKey02Code    :
      break ;
    case LongKey03Code    :
      break ;
    case LongKey04Code    :
      break ;
    case LongKey05Code    :
      break ;
    case LongKey06Code    :
      break ;
    case LongKey07Code    :
      break ;
    case LongKey12Code    :
      break ;
    case LongKey13Code    :
      break ;
    case LongKey14Code    :
      break ;
    case LongKey15Code    :
      break ;
    case LongKey16Code    :
      break ;
    case LongKey17Code    :
      break ;
    case LongKey23Code    :
      break ;
    case LongKey24Code    :
      break ;
    case LongKey25Code    :
      break ;
    case LongKey26Code    :
      break ;
    case LongKey27Code    :
      break ;
    case LongKey34Code    :
      break ;
    case LongKey35Code    :
      break ;
    case LongKey36Code    :
      break ;
    case LongKey37Code    :
      break ;
    case LongKey45Code    :
      break ;
    case LongKey46Code    :
      break ;
    case LongKey47Code    :
      break ;
    case LongKey56Code    :
      break ;
    case LongKey57Code    :
      break ;
    case LongKey67Code    :
      break ;
  } /* switch */ ;
} /* ExecuteKeyCommand */


static uint8_t SelectStartDevIndex ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < ChannelsArraySize ; Loop ++ )
  {
    if ( EEBlock . SubDeviceID [ Loop ] [ 0 ] )
    {
      return Loop ;
    } /* if */ ;
  } /* for */ ;
  return NoActive ;
} /* SelectStartDevIndex */


static uint8_t SelectNextDevice ( uint8_t StartIndex )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( StartIndex == NoActive )
    return NoActive ;
  StartIndex ++ ;
  if ( StartIndex >= ChannelsArraySize )
    StartIndex = 0 ;
  for ( Loop = StartIndex ; Loop < ChannelsArraySize ; Loop ++ )
  {
    if ( EEBlock . SubDeviceID [ Loop ] [ 0 ] )
    {
      return Loop ;
    } /* if */ ;
  } /* for */ ;
  return SelectStartDevIndex ( ) ;
} /* SelectNextDevice */


static uint8_t SendReplyReq ( uint8_t DevIndex )
{
  /*----------------------------------------------------------------*/
  if ( DevIndex != 0xFF )
  {
    SendTxt ( & SlaveDevService , EEBlock . SubDeviceID [ DevIndex ] ) ;
    Send ( & SlaveDevService , PartSeparator ) ;
    SendTxt ( & SlaveDevService , EEBlock . MyDeviceID ) ;
    Send ( & SlaveDevService , PartSeparator ) ;
    SendFlashTxt ( & SlaveDevService , ( uint16_t ) ReplyReqCommand ) ;
    Send ( & SlaveDevService , cr ) ;
    return true ;
  } /* if */ ;
  return false ;
} /* SendReplyReq */


static void ResetOperation ( PtrSideServiceType SideData )
{
  /*----------------------------------------------------------------*/
  ClearBlock ( SideData -> CommandText , sizeof ( CommandBufferType ) ) ;
  ClearBlock ( SideData -> ParamText , sizeof ( ParamBufferType ) ) ;
  ClearBlock ( SideData -> DestinationDeviceName , sizeof ( DeviceIDType ) ) ;
  ClearBlock ( SideData -> SourceDeviceName , sizeof ( DeviceIDType ) ) ;
  SideData -> CommandTextIndex = 0 ;
  SideData -> ParamTextIndex = 0 ;
  SideData -> DestDeviceNameIndex = 0 ;
  SideData -> SrcDeviceNameIndex = 0 ;
  SideData -> SerialService = SelectDestOperation ;
  SideData -> CommandBuffer . CommandLineIndex = 0 ;
} /* ResetOperation */


static void NoEcho ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
} /* NoEcho */


static void SoftwareInit ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  ConnectContext = false ;
  CurrentTimeStatusState = 0 ;
  USART232Reset ( ) ;
  USART485Reset ( ) ;
  SuperVisService . SendSerial = USART232SendSerial ;
  SuperVisService . EchoSerial = NoEcho ;
  SuperVisService . SerialDataPresent = USART232DataPresent ;
  SuperVisService . GetSerial = USART232GetSerial ;
  SuperVisService . ExecuteCommand = SuperVisExecuteCommand ;
  SlaveDevService . SendSerial = USART485SendSerial ;
  SlaveDevService . EchoSerial = NoEcho ;
  SlaveDevService . SerialDataPresent = USART485DataPresent ;
  SlaveDevService . GetSerial = USART485GetSerial ;
  SlaveDevService . ExecuteCommand = SubDevExecuteCommand ;
  TimerCounter = 0 ;
  ResetOperation ( & SuperVisService ) ;
  ResetOperation ( & SlaveDevService ) ;
  for ( Loop = 0 ; Loop < DisplParamArraySize ; Loop ++ )
  {
    ClearBlock ( & ParamValueArray [ Loop ] , sizeof ( ParamValueDescrRecT ) ) ;
  } /* for */ ;
  KeybStatus = KeybIdle ;
  KeybCode = 0 ;
  KeybDelayCt = 0 ;
  KeybDelayTimer = 0 ;
  LEDModeSwitch = StatusDispMode ;
  SubDevTimerCounter = 0 ;
  AnyKeyDelay = 0 ;
  Events = 0 ;
  CurrentFunctionSelected = Function0Selected ;
  QueueTimerCounter = 0 ;
} /* SoftwareInit */


int main ( void )
{
  /*----------------------------------------------------------------*/
  HardwareInit ( Speed38400bps , Speed9600bps ) ;
  EnvirInit ( ) ;
  SoftwareInit ( ) ;
  EEBlockOperation ( ) ;
  LEDModeSwitch = EEBlock . DisplayMode ;
  USART485ReceiveMode ( ) ;
  wdt_enable ( WDTO_1S ) ;
  SubdeviceIndex = SelectStartDevIndex ( ) ;
  sei ( ) ;
  SetFunctionIndicator ( CurrentFunctionSelected ) ;
  LinkParamValues ( ) ;
  ClearAllScreens ( ) ;
  if ( EEBlock . StartKeyCode != NoActive )
  {
    ExecuteKeyCommand ( EEBlock . StartKeyCode ) ;
  } /* if */ ;
  for ( ; ; )
  {
    wdt_reset ( ) ;
    if ( Events & TimeEvent )
    {
      Events &= ~ TimeEvent ;
      if ( CurrentTimeStatusState )
        LifeLEDPort &= ~ ( ( 1 ) << LifeLED ) ;
      else
        LifeLEDPort |= ( 1 ) << LifeLED ;
      CurrentTimeStatusState = 1 - CurrentTimeStatusState ;
    } /* if */ ;
    if ( Events & SubDeviceEvent )
    {
      Events &= ~ SubDeviceEvent ;
      if ( SendReplyReq ( SubdeviceIndex ) )
        SubdeviceIndex = SelectNextDevice ( SubdeviceIndex ) ;
    } /* if */ ;
    if ( Events & QueueEvent )
    {
      Events &= ~ QueueEvent ;
      SubDevFlushData ( ) ;
    } /* if */ ;
    if ( KeyPressed ( ) )
    {
      ExecuteKeyCommand ( ReadKey ( ) ) ;
    } /* if */ ;
    if ( SuperVisService . SerialService )
      SuperVisService . SerialService ( & SuperVisService ) ;
    if ( SlaveDevService . SerialService )
      SlaveDevService . SerialService ( & SlaveDevService ) ;
  } /* for */ ;
  return ( 0 ) ;
} /* main */


